package biss;

import java.util.Enumeration;
import java.util.NoSuchElementException;

/**
 * This is an abstract base for Hashtabs with primitive keys (long, int, ..).
 * Such Hashtabs are justified only by performance and memory considerations. There
 * is not much to factor out because (of the above reasons) the interesting methods
 * in the subclasses (indexOf..) have to deal explicitly with the key type. Getting
 * more functionality in the base would be possible (by means of additional methods),
 * but would also sacrifice speed. This class wouldn't exist if it wouldn't be
 * twice as fast, consuming just half of the memory in case of biss.IntHashtab vs.
 * java.util.Hashtable (with Integer keys).
 *
 * (C) 1997 BISS GmbH Germany, see file 'LICENSE.BISS-AWT' for details
 * @author P.C.Mehlitz
 */
abstract public class PrimHashtab
{
	protected Object[] Values;
	protected Object KeyArray;
	protected int Count;
	protected int LeftShift;
	protected Object RecycledEntry;

protected PrimHashtab () {
}

public PrimHashtab ( int entryGuess, int leftShift ) {
	int n = (entryGuess > 20) ? getSizePrime( entryGuess * 3 / 2) : 17;
	
	Values    = createValueArray( n);
	KeyArray  = createKeyArray( n);
	RecycledEntry = createRecycledEntry();

	LeftShift = leftShift;
}

public synchronized void clear () {
	int n = Values.length;

	Count    = 0;
	Values   = new Object[n];
	KeyArray = createKeyArray(n);
}

abstract protected Object createKeyArray ( int n );

protected Object createRecycledEntry () {
	return new Integer(0);
}

protected Object[] createValueArray ( int n ) {
	return new Object[n];
}

public synchronized Enumeration elements () {
	return new PrimHashtabEnumerator( Values, RecycledEntry);
}

static int getSizePrime ( int n ) {
	if ( n > 10000 )     return 14983;
	else if ( n > 5000 ) return 7507;
	else if ( n > 2000 ) return 2503;
	else if ( n > 1000 ) return 1511;
	else if ( n >  800 ) return  907;
	else if ( n >  600 ) return  701;
	else if ( n >  500 ) return  557;
	else if ( n >  400 ) return  457;
	else if ( n >  300 ) return  353;
	else if ( n >  200 ) return  257;
	else if ( n >  100 ) return  151;
	else if ( n >   80 ) return   89;
	else if ( n >   60 ) return   71;
	else if ( n >   40 ) return   53;
	else if ( n >   20 ) return   31;
	else                 return   13;
}

public boolean isEmpty () {
	return (Count == 0);
}

public int size() {
	return Count;
}
}

class PrimHashtabEnumerator
  implements Enumeration
{
	Object[] Values;
	int Idx;
	Object NextElement;
	Object RecycledEntry;

PrimHashtabEnumerator ( Object[] values, Object recycledEntry ) {
	Values = values;
	Idx = -1;
	RecycledEntry = recycledEntry;
}

public boolean hasMoreElements () {
	Object v = null;
	int    n = Values.length;

	if ( NextElement == null ) {
		for ( Idx++; Idx < n && ((v=Values[Idx]) == null || v == RecycledEntry); Idx++ );

		if ( Idx < n ) {
			NextElement = v;
			return true;
		}
		else {
			NextElement = null;
			return false;
		}
	}
	else
		return true;
}

/**
 * Concrete iterator for PrimHashtabs. Please note that java.util.Enumeration
 * implementors should always be robust, i.e. if "hasMoreElements()" returns
 * true, a subsequent "nextElement()" should never cause an exception (regardless
 * of which operations are performed on the container between these two method
 * invocations). However, this does not mean that an Iteration visits all
 * container elements if Iteration is done overlapping with put() / remove()
 * operations.
 */
public Object nextElement () {
	Object v = NextElement;
	
	if ( v == null ) {
		if ( hasMoreElements() )
			v = NextElement;
		else
			throw new NoSuchElementException("PrimHashtabEnumerator");
	}
	
	NextElement = null;
	return v;
}
}
