package biss;

import biss.PrimHashtab;
import biss.VectorLib;
import biss.VectorSorter;
import java.util.Enumeration;
import java.util.Vector;

/**
 * Open hashing based set implementation.
 *
 * (C) 1997 BISS GmbH Germany, see file 'LICENSE.BISS-AWT' for details
 * @author P.C.Mehlitz
 */
public class Set
  extends PrimHashtab
  implements Cloneable
{
public Set () {
	super( 17, 0);
}

public Set ( Object[] a ) {
	super( a.length, 0);
	
	addAll( a);
}

public Set ( Set set ) {
	Values = (Object[]) set.Values.clone();
	Count = set.Count;
}

public Set ( Vector v ) {
	super( v.size(), 0);
	
	addAll( v.elements());
}

public Set ( int entryGuess ) {
	super( entryGuess, 0);
}

public synchronized Object add ( Object obj ) {
	int     i;
	Object  v;

	do {
		if ( (i = indexOf( obj, true)) >= 0 ) {
			if ( ((v = Values[i]) == null) || (v == RecycledEntry) ) {
				Values[i] = obj;
				Count++;
				return null;
			}
			else
				return v;
		}
	}
	while ( grow() ); // no space available any more
	
	throw new NullPointerException();
}

public synchronized void addAll ( Enumeration e ) {
	for ( ; e.hasMoreElements(); )
		add( e.nextElement());
}

public synchronized void addAll ( Object[] a ) {
	for ( int i=0; i<a.length; i++ )
		add( a[i]);
}

public Object clone () {
	Set set;
	
	try {
		set = (Set) super.clone();
		set.Values = (Object[]) Values.clone();
		
		return set;
	}
	catch ( CloneNotSupportedException x ) {
		return null;
	}
}

final public boolean contains ( Object obj ) {
	return indexOf( obj, false) >= 0;
}

public boolean containsAll ( Enumeration e ) {
	for ( ; e.hasMoreElements(); ) {
		if ( !contains( e.nextElement()) )
			return false;
	}
	
	return true;
}

public boolean containsAll ( Object[] a ) {
	for ( int i=0; i<a.length; i++ ) {
		if ( !contains( a[i]) )
			return false;
	}
	
	return true;
}

public boolean containsAny ( Enumeration e ) {
	for ( ; e.hasMoreElements(); ) {
		if ( contains( e.nextElement()) )
			return true;
	}
	
	return false;
}

public boolean containsAny ( Object[] a ) {
	for ( int i=0; i<a.length; i++ ) {
		if ( contains( a[i]) )
			return true;
	}
	
	return false;
}

public boolean containsNone ( Enumeration e ) {
	for ( ; e.hasMoreElements(); ) {
		if ( contains( e.nextElement()) )
			return false;
	}
	
	return true;
}

public boolean containsNone ( Object[] a ) {
	for ( int i=0; i<a.length; i++ ) {
		if ( contains( a[i]) )
			return false;
	}
	
	return true;
}

protected Object createKeyArray ( int n ) {
	return null;
}

boolean grow () {
	int m = Values.length;
	int n = m * 3 / 2;
	int i, k;
	Object v;

	if ( m > 14983 ) return false;

	Set newSet = new Set( n);
	
	for ( i=0, k=Count; i<m; i++ ) {
		if ( (v = Values[i]) != null && v != RecycledEntry ) {
			newSet.add( v);
			if ( --k == 0 ) break;
		}
	}
	
	Values = newSet.Values;
	
	return true;
}

final private int indexOf ( Object obj, boolean newOne ) {
	int    k = obj.hashCode();
	int    m = Values.length, m2 = m >> 1;
	int    i, j = 0, d = 1;
	Object v;

	if ( k < 0 ) k = -k;
	i = (k % m);

  do {
		if ( (v = Values[i]) == null ){
			return (newOne) ? i : -1;
		}
		else if ( v == RecycledEntry ) {
			if ( newOne ) return i;
		}
		else if ( v.equals( obj) ) {
			return i;
		}
		
		i += d; d += 2;
		if ( i >= m )
			i %= m;
	}
	while ( ++j <= m2 );
	
	return -1;
}

public static void main ( String[] args ) {
	Set s = new Set();
	
	s.add( "eins");
	s.add( new String( "eins") );
	s.add( "zwei");
	s.add( "zwei");
	s.add( "zwoi");
		System.out.println( s.toVector());
	
	System.out.println( "remove zwickel: " + s.remove( "zwickel"));
	System.out.println( "remove eins:    " + s.remove( "eins"));
	System.out.println( s.toVector());
	
}

public synchronized Object remove ( Object obj ) {
	int i = indexOf( obj, false);

	if ( i >= 0 ) {
		Object v = Values[i];
		Values[i] = RecycledEntry;
		Count--;
		return v;
	}
	else
		return null;
}

public synchronized void removeAll ( Enumeration e ) {
	for ( ; e.hasMoreElements(); )
		remove( e.nextElement());
}

public synchronized void removeAll ( Object[] a ) {
	for ( int i=0; i<a.length; i++ )
		remove( a[i]);
}

public synchronized Object[] toArray () {
	int i, j, n = Values.length;
	Object[] a = new Object[Count];
	Object v;
	
	for ( i=0, j=0; i<n; i++ ) {
		v = Values[i];
		if ( v != null && v != RecycledEntry )
			a[j++] = v;
	}
	
	return a;
}

public Vector toSortedVector () {
	return toSortedVector( null);
}

public synchronized Vector toSortedVector ( VectorSorter sorter ) {
	int i, n = Values.length;
	Vector list = new Vector( Count);
	Object v;
	
	for ( i=0; i<n; i++ ) {
		v = Values[i];
		if ( v != null && v != RecycledEntry )
			VectorLib.sortIn( list, v, sorter);
	}
	
	return list;    
}

public synchronized Vector toVector () {
	int i, n = Values.length;
	Vector list = new Vector( Count);
	Object v;
	
	for ( i=0; i<n; i++ ) {
		v = Values[i];
		if ( v != null && v != RecycledEntry )
			list.addElement( v);
	}
	
	return list;    
}

public Set union ( Enumeration e ) {
	Set s = new Set();
	for ( ; e.hasMoreElements(); ) {
		Object o = e.nextElement();
		if ( contains( o) )
			s.add (o);
	}
	
	return s;
}
}
