package biss.awt;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Event;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.MenuBar;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * replacement for java.awt.MenuBar
 *
 * (C) 1996,97 BISS GmbH Germany, see file 'LICENSE.BISS-AWT' for details
 * @author J.H.Mehlitz
 */
public class BarMenu
  extends Widget
  implements HotKeyProcessor, MouseMotionInterface
{
	MenuItem Selection;
	FontMetrics Fm;
	MenuItemContainer Contents;
	int Dx;

public BarMenu() {
	setForeground( Awt.ButtonForeClr);
	setBackground( Awt.ButtonBackClr);
	setCursor( Cursor.getPredefinedCursor( Awt.ActionCursor));
	setFont( Awt.MenuFont);
	
	FBorder.setBorder( Border.BS_OUTSET);
	registerHotKeys();
}

public BarMenu( MenuItemContainer mic) {
	this();
	setContents( mic);
}

public BarMenu( String items) {
	this();
	setContents( new Menu("", items));
}

public synchronized void addNotify () {
	super.addNotify();
	registerKeys();
}

protected boolean changeSelection( MenuItem newSel ){
	if ( newSel == Selection)
		return false;

	MenuItem oldSel = Selection;
	Selection = null;

	if ( oldSel != null) {
		if ( oldSel instanceof MenuItemContainer ){
			EventHandler.disposeSingleInstWnd();
			oldSel.setExpanded( false);
		}
		redrawItem( oldSel);
	}
	
	Selection = newSel;			
	if ( Selection != null)
		redrawItem( Selection);

	return true;
}

protected boolean cursorLeft( ){
	if ( (Selection == null) || (Contents == null) )
		return false;

	Vector v = Contents.getItems();
	int fsi = v.indexOf( Selection);
	if ( fsi < 1)
		return false;

	return changeSelection( (MenuItem)v.elementAt( fsi - 1));
}

protected boolean cursorRight( ){
	if ( (Selection == null) || (Contents == null) )
		return false;
		
	Vector v = Contents.getItems();
	int fsi = v.indexOf( Selection);
	if ( fsi >= v.size() - 1)
		return false;

	return changeSelection( (MenuItem)v.elementAt( fsi + 1));
}

public MenuItemContainer getContents(){
	return Contents;
}

public static int getDefaultHeight() {
	return (3 * Awt.SysFontHeight) / 2;
}

public Vector getItems(){
	return (Contents != null) ? Contents.getItems() : null;
}

public Menu getMenu(){
	if ( (Contents != null) && (Contents instanceof Menu))
		return (Menu)Contents;
	return null;
}

public void gotFocus(){
	if ( Selection == null)
		Selection = itemAtPos( 0);
	redrawItem( Selection);
}

protected MenuItem itemAtPos( int xPos){
	if ( Contents == null)
		return null;
		
	int p = FBorder.Ext;
	Vector v = Contents.getItems();
	int    vs = v.size();
	
	for ( int i=0; i<vs; i++){
		MenuItem mi = (MenuItem)v.elementAt( i);
		p += Fm.stringWidth( mi.Text) + 2*Dx;
		if ( p > xPos )
			return mi;
	}
	return null;
}

public void lostFocus(){
	if ( Selection != null) {
		MenuItem lastSel = Selection;
		Selection = null;
		redrawItem( lastSel);
	}
}

public void mouse1Down(){
	updateSelection();
}

public void mouse1Drag(){
	updateSelection();
}

public void mouse1Up(){
	if ( Selection != null)
		processSelection();
}

public void mouse2Drag(){
}

public void mouseEnter(){
	updateSelection();
}

public void mouseExit(){
	if ( EventHandler.HasFocus)
		return;
	if ( (Selection != null) && ( ! Selection.isExpanded()) )
		changeSelection( null);
}

public void mouseMove(){
//	if ( EventHandler.HasFocus)
		updateSelection();
}

public void processKey( Object sel){
	if ( "cursorLeft".equals( sel) )
		cursorLeft();
	else if ( "cursorRight".equals( sel) )
		cursorRight();
	else if ( "processFocus".equals( sel) )
		processSelection();
	else
		trigger( (MenuItem)sel );
}

boolean processSelection( ){
	if ( (Selection == null) || !( Selection.isEnabled()) )
		return false;

	redrawItem( Selection);
	
	EventHandler.disposeSingleInstWnd();
	Selection.setExpanded( false);
	
	if (Selection instanceof MenuItemContainer){
		Vector v = ((MenuItemContainer)Selection).getItems();
		if ( (v != null) && (v.size() > 0) ){
			PopupMenu p = new PopupMenu( this, v);
			Point pt = Awt.getDesktopLocation( this, xPosFor( Selection), Height);
			p.popUpAt( pt.x, pt.y);
			Selection.setExpanded( true);
			return true;
		}
	}

	trigger();
	return true;
}

public void redraw( Graphics g){
	int y = Height - ( Height - Fm.getHeight()) / 2 - (Fm.getDescent()+2); 
	int x = FBorder.Ext;

	blank( g);
	clip( g);

	if ( Contents == null)
		return;
		
	g.setFont( getFont() );

	Vector v = Contents.getItems();
	int    vs = v.size();

	for ( int i=0; i<vs; i++) {
		MenuItem mi = (MenuItem)v.elementAt( i);
		mi.drawAt( g, x+Dx, y, false);
		if ( EventHandler.HasFocus && (mi == Selection))
			GraphicsLib.drawFocusMark( g, x, FBorder.Ext, 4);
		x += Fm.stringWidth( mi.Text)+2*Dx;
	}    
}

protected void redrawItem( MenuItem mi){
	redrawItem( mi, false);
}

protected void redrawItem( MenuItem mi, boolean flash){

	if ( mi == null)
		return;

	int y = Height - ( Height - Fm.getHeight()) / 2 - (Fm.getDescent()+2); 
	int x = xPosFor( mi);
	int mw = Fm.stringWidth( mi.Text);

	Graphics g = getGraphics();
	
	if ( g == null)
		return;

	g.setColor( getBackground() );
	g.fillRect( x, DrawRect.y, mw+2*Dx, DrawRect.height);
	g.setFont( getFont() );

	mi.drawAt( g, x+Dx, y, flash);
//	if ( EventHandler.HasFocus && (mi == Selection))
	if ( mi == Selection)
		GraphicsLib.drawFocusMark( g, x, FBorder.Ext, 4);
		
	g.dispose();
}

protected void registerHotKeys(){
	addHotKey( Key.CursorLeft, HotKey.Lone, this, "cursorLeft");
	addHotKey( Key.CursorRight, HotKey.Lone, this, "cursorRight");
	addHotKey( Key.Newline, HotKey.Lone, this, "processFocus");
}

protected boolean registerKeys(){
	Frame fr = getFrame();
	GUIHandlerInterface gi = ( fr instanceof GUIHandlerInterface) ?
														 (GUIHandlerInterface) fr : null;

	if ( (gi == null) || (Contents == null) )
		return false;

	gi.getGUIHandler().addHotKeys( Contents.getHotKeys( this, Awt.GlobalKeyMod));
	return true;
}

public void setContents( MenuItemContainer mic){
	Contents = mic;
	updateContents();
}

public void setEnabled( boolean b){
	Contents.setEnabled( b);
	super.setEnabled( b);
}

public synchronized void setFont( Font fnt) {
	Font uf = ( fnt != null) ? fnt : Awt.MenuFont;

	super.setFont( uf);
	Fm = Awt.DefToolkit.getFontMetrics( uf);
	Dx = 2*Fm.charWidth( 'x')/3;

	if ( isShowing() )
		repaint();
}

public void setMenu( Menu m){
	setContents( m);
}

protected boolean trigger(){
	return trigger( Selection);
}

protected boolean trigger( MenuItem sel){
	if ( (sel == null) || !sel.isEnabled() )
		return false;

	redrawItem( sel, true);
	Awt.DefToolkit.sync();

	// invoke java.awt.MenuItem listeners
	sel.process();	//may remove selection
			
	if ( sel != null)
		OsCommand.notifyObservers( sel.Selector);
			
	try { Thread.sleep( 200);}
	catch( Exception e) {}

	redrawItem( sel, false);
		
	return true;
}

public void updateContents( ){
	if ( isShowing() )
		redraw();
}

protected void updateSelection(){
	changeSelection( itemAtPos( Mouse.Pos.x));
}

protected int xPosFor( MenuItem mi){
	if ( Contents == null)
		return 0;

	int p = FBorder.Ext;
	Vector v = Contents.getItems();
	int    vs = v.size();
	
	for ( int i=0; i<vs; i++) {
		MenuItem ci = (MenuItem)v.elementAt( i);
		if ( ci == mi)
			return p;
		p += Fm.stringWidth( ci.Text)+2*Dx;
	}
	return 0;
}
}
