package biss.awt;

import biss.CommandQueue;
import biss.ObserverSocket;
import biss.PackageProperties;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Window;
import java.awt.event.*;
import java.util.Observable;
import java.util.Observer;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * replaces java.awt.Choice
 *
 * (C) 1996,97 BISS GmbH Germany, see file 'LICENSE.BISS-AWT' for details
 * @author J.H.Mehlitz
 */
public class Choice
  extends Compound
  implements FocusSetInterface, HotKeyProcessor, Observer, MouseListener
{
	final public static int READ_WRITE = 1;
	final public static int READ_ONLY = 0;
	protected ChoiceWindow Cw;
	protected Vector Data;
	protected TextField Entry = new TextField();
	protected int Type = 1;
/**
 * triggered on selection of a choice item
 * parameter: selection
 */
	public ObserverSocket OsSelection = new ObserverSocket( this);
	protected Object Selection;
/**
 * triggered by a choice item paint request , just in case
 * of at least one registered observer.
 * This observer is responsible for all item painting
 * ( overrides the standard string representation)
 * parameter: choice item
 */
	public ObserverSocket OsDrawObject = new ObserverSocket( this);
	protected long LastDisposeTime;

public Choice(){
	FocusForward = Entry;
	Entry.setBorder( Border.BS_NONE);
	FBorder.setBorder( TextField.DefaultBorderStyle);

	setLayout( null);
	add( Entry);
	
	setCursor( Cursor.getPredefinedCursor( Awt.ActionCursor));
	registerObserver();
	registerHotKeys();
	addMouseListener( this);
}

public Choice( String items){
	this();
	Data = new Vector();
	StringTokenizer tok = new StringTokenizer( items, ",:;");
	while( tok.hasMoreTokens() )
		Data.addElement( tok.nextToken() );
}

public Choice( Vector data){
	this();
	Data = data;
}

public Choice( int type, Vector data) {
	this();
	Type = type;
	Data = data;
	Entry.setReadOnly( Type == READ_ONLY);
}

public void appendItem (Object obj){
	if ( Data == null)
		Data = new Vector();
	Data.addElement( obj);
	
	if ( Selection == null)
		setSelection( 0);
}

public boolean appendItemUnique (Object obj){
	if ( (Data != null) && (Data.contains( obj)) )
		return false;
	appendItem( obj);
	return true;
}

void buildChoice(){
	Cw = new ChoiceWindow( this);
	Cw.DataPane.OsDrawObject.addObserver( this);
}

public boolean canGetFocus(){
	return Entry.canGetFocus();
}

public void clearSelection (){
	Selection = null;
	Entry.setContents( "");
}

void closeChoice(){
	if ( Cw != null ) {
		GUIHandler.SingleInstWnd = null;
		Cw.dispose();
		Cw = null;
		redraw();
		Entry.requestFocus();
		LastDisposeTime = System.currentTimeMillis();
	}
}

public void doLayout(){
	int bw = FBorder.Ext;
	Entry.setBounds( bw, bw, Width-Awt.ScrollBarWidth-2-2*bw, Height-2*bw );
}

public Vector getContents(){
	return Data;
}

public TextField getEntry(){
	return Entry;
}

public Object getItem ( int index){
	try { return Data.elementAt( index); }
	catch( Throwable t) { return null; }
}

public int getItemCount (){
	try { return Data.size(); }
	catch ( Throwable t) { return 0; }
}

public List getList(){
	if ( Cw != null)
		return Cw.DataPane;
	return null;
}

public Dimension getMinimumSize(){
	int maxDx = 0;
	FontMetrics fm = getFontMetrics( Entry.getFont() );
	if ( Data != null) {
		for ( int i=0; i<Data.size(); i++){
			Object o = Data.elementAt( i);
			if ( o instanceof String){
				int dx = fm.stringWidth( (String)o);
				if ( dx > maxDx)
					maxDx = dx;
			}
		}
	}
	
	return new Dimension( 30 + maxDx, 10 + fm.getHeight() );
//	return new Dimension( 10*Awt.SysFontWidth, 3*Awt.SysFontHeight/2);
}

void getProcessed ( Object sel){
	// null selection means click in blank ChoiceWindow area
	// do not change contents or notifys
	if ( (sel != null) && setSelection( sel) )
		OsSelection.notifyObservers( Selection);
	closeChoice();
}

public Object getSelection (){
	return Selection;
}

public int getSelectionIdx (){
	if ( (Data == null) || (Selection == null) )
		return -1;
	return Data.indexOf( Selection);
}

public boolean hasFocus(){
	return Entry.hasFocus();
}

public boolean insertItemBefore (Object obj, int index){
	if ( Data == null)
		Data = new Vector();
	try {
		Data.insertElementAt( obj, index);
		return true;
	}
	catch ( Throwable t) {
		return false;
	}
}

boolean isChoiceOpen() {
	return ( (Cw != null) && (Cw.isShowing()));
}

public void mouseClicked( MouseEvent e) {
	e.consume();
}

public void mouseEntered( MouseEvent e) {
	if ( Enabled)
		setCursor();
	e.consume();
}

public void mouseExited( MouseEvent e) {
	e.consume();
}

public void mousePressed( MouseEvent e) {
	if ( Enabled) {
		// Windows workaround to toggle choice button
		if ( System.currentTimeMillis() - LastDisposeTime < 500 )
			return;
		if ( isChoiceOpen() )
			closeChoice();
		else
			openChoice();
	}
	e.consume();
}

public void mouseReleased( MouseEvent e) {
	e.consume();
}

void openChoice() {
	Point pt = Awt.getDesktopLocation( Entry, 0, Entry.Height);

	GUIHandler.disposeSingleInstWnd( this);
	if ( Cw == null)
		buildChoice();

	//perform List default drawing when Drawsocket is empty
	Cw.DataPane.overrideDrawSocket( OsDrawObject.isEmpty() );  

	Cw.popUpAt( pt.x - FBorder.Ext, pt.y + FBorder.Ext, Width);
	Cw.initLabel( Entry.getContents());

	//draw pushed button
	redraw();
	
	CommandQueue.append( new ToFrontCmd( Cw));
	CommandQueue.append( new SetFocusCmd( Cw.DataPane.Cv));
	GUIHandler.SingleInstWnd = Cw;
	
}

public void processKey(Object sel){
	if ( sel.equals( "popDown"))
		openChoice();
}

public void redraw( Graphics g){

	int x = Entry.Width + FBorder.Ext;
	int y = FBorder.Ext - 1;
	int w = Awt.ScrollBarWidth + 2;
	int h = Entry.Height + 1;

	GraphicsLib.paintChoiceButton( g, x, y, w, h, Cw != null);
}

void registerHotKeys (){
	Entry.addHotKey( Awt.ChoiceOpenKeys[0], Awt.ChoiceOpenKeys[1], this, "popDown");
}

void registerObserver (){
	Entry.OsFocusGot.addObserver( this);
	Entry.OsFocusLost.addObserver( this);
	Entry.OsEnter.addObserver( this);
}

public void removeAllItems(){
	clearSelection();
	Data = null;
}

public boolean removeItem (Object obj){
	try {
		Data.removeElement( obj);
		if ( obj == Selection)
			clearSelection();
		return true;
	}
	catch( Throwable t) {
		return false;
	}
}

public boolean removeItem ( int index){
	return removeItem( getItem( index));
}

public synchronized void setBackground( Color c){
	Entry.setBackground( c);

	Entry.redraw(); // Windows bug
}

public void setContents( Vector v){
	Data = v;
	setSelection( 0);
}

public boolean setFocus(){
	return Entry.setFocus();
}

public boolean setFocusIfActive () {
	TopWindow tw = (TopWindow) getFrame();

	if ( tw != null && tw.isActive() )
		return setFocus();
	else
		return false;
}

public synchronized void setFont( Font f){
	Entry.setFont( f);

	Entry.redraw(); // Windows bug
}

public synchronized void setForeground( Color c){
	Entry.setForeground( c);

	Entry.redraw(); // Windows bug
}

public void setReadOnly( boolean state){
	Entry.setReadOnly( state);
}

public boolean setSelection ( Object sel){
	if ( sel == Selection)
		return false;
		
	if ( sel == null){
		clearSelection();
		return true;
	}
	
	if ( Type == READ_ONLY){
		if ( (Data == null) || (!Data.contains( sel)) )
			return false;
	}
	
	Selection = sel;
	if ( sel instanceof SelfDrawingObject)
		Entry.setContents( ((SelfDrawingObject)sel).label() );
	else
		Entry.setContents( sel.toString() );

	return true;
}

public boolean setSelection ( int index){
	return setSelection( getItem( index) );
}

public String stringSelection (){
	Selection = Entry.getContents();
	return (String)Selection;
}

/**
 *
 */
public void update (Observable obs, Object arg){
	if ( obs == Entry.OsFocusGot)
		OsFocusGot.notifyObservers();
	else if ( obs == Entry.OsFocusLost)
		OsFocusLost.notifyObservers();
	else if ( obs == Entry.OsEnter){
		Selection = Entry.getContents();
		OsSelection.notifyObservers( Selection);
	}
	else if ( (Cw != null) && (obs == Cw.DataPane.OsDrawObject) )
		OsDrawObject.notifyObservers( arg);
}
}

class ChoiceWindow
  extends Window
  implements Observer, SingleInstInterface
{
	Choice CP = null;
	List DataPane = new List( List.SB_VERT);
	Layouter Lo = new Layouter( this);

public ChoiceWindow ( Choice cp) {
	super( cp.Entry.getFrame());
	CP = cp;
	Lo.add( DataPane, 0, 0, 1000, 1000, false, 2, 2, -2, -2, 0);
	
	// *** OS/2 workaround ***
	if ( Awt.AwtToolkit == Awt.TK_OS2)
		DataPane.Cv.UseCopyArea = false;
		
	registerObserver();
	
}

public boolean disposeRequest( Component nextFocus){
	if ( (nextFocus != DataPane.Cv) && (nextFocus != DataPane.Vert)  ) {
		CP.closeChoice();
		return true;
	}
	return false;
}

public boolean initLabel( String label){
	return ((ListCanvas)DataPane.BaseCv).selectLabel( label);
}

public void paint ( Graphics g){
	super.paint( g);
	Rectangle r = getBounds();
	g.setColor( Color.black);
	g.drawRect( 0, 0, r.width-1, r.height-1);
}

public void popUpAt( int x, int y, int width){
	DataPane.setForeground( CP.Entry.getForeground() );
	DataPane.setBackground( CP.Entry.getBackground() );

	setBounds( x, y, width, 4*Awt.SysFontHeight );
	setVisible( true);
//	toFront();
	DataPane.setContents( CP.Data);
	DataPane.Cv.requestFocus();
}

public void registerObserver (){
	DataPane.OsProcess.addObserver( this);
	DataPane.OsSelection.addObserver( this);
}

public void update ( Observable o, Object arg){
	if ( DataPane.OsProcess.equals( o))
		CP.getProcessed( DataPane.getSelection() );
	else if ( DataPane.OsSelection.equals( o)) {
		if ( DataPane.mouseSelection() )
			CP.getProcessed( DataPane.getSelection() );
	}
}
}
