package biss.awt;

import java.util.Vector;

/**
 * wrapper object for hierarchically organized data (e.g. a single directory
 * object). Stores display state and Level information. LeveledObjects
 * are usually created by Levelers. Used by LeveledList.
 *
 * Note that LeveledObjects are created implicitly via the static add(..)
 * method and there is no public ctor. This is because LeveledObjects are
 * context sensitive (live within a Vector of LeveledObjects). It would be
 * error prone to force all applications to check for leveled list consistency
 *
 * (C) 1996,97 BISS GmbH Germany, see file 'LICENSE.BISS-AWT' for details
 * @author J.H.Mehlitz
 */
public class LeveledObject
{
	final public static int LO_EXPANDABLE = 1;
	final public static int LO_EXPANDED = 2;
	final public static int LO_VISIBLE = 4;
	int Level;
	Object Obj;
	int State;
	long FollowLevels;

LeveledObject ( Object obj, int lvl) {
	Obj   = obj;
	Level = lvl;
	State = 0;
}

public static void add (Vector v, Object obj, int lvl) {
	int iLvl = 0;

	try {
		LeveledObject llo = (LeveledObject)v.lastElement();
		iLvl = Math.min( llo.Level+1, lvl);
	}
	catch ( Throwable t) {}

	LeveledObject nlo = new LeveledObject( obj, iLvl);
	int s = v.size();
	int ll = iLvl;

	if ( iLvl == 0)
		nlo.setVisible( true);

	for ( int i=s-1; i>-1; i--){
		LeveledObject lo = (LeveledObject)v.elementAt( i);
		if ( lo.Level >= ll)
			break;
		lo.setExpandable( true);
		ll = lo.Level;
	}

	v.addElement( nlo);
}

public void addState (int flags) {
	State |= flags;
}

public static void computeFollowLevels ( Vector v ) {
	int n = v.size();
	LeveledObject lo, lo1;
	int i, j, k, lvl;
	long follow;

	for ( i=0; i<n; i++ ) {
		lo1 = (LeveledObject)v.elementAt( i);
		lvl = lo1.Level;
		
		for ( follow = 1, j=0; j<lvl; j++ )
			follow = (follow << 1) | 1;

		for ( k = lvl; k >= 0; k-- ) {
			for ( j=i+1; j<n; j++ ) {
				lo = (LeveledObject)v.elementAt( j);
				lvl = lo.Level;
				if ( lvl < k ) {
					follow &= ~(1 << k);
					break;
				}
				else if ( lvl == k ) {
					break;
				}
			}
			if ( j == n )
				follow &= ~(1 << k);					
		}
		
		lo1.FollowLevels = follow;
	}
}

public static void expandAll ( Vector v ) {
	expandToLevel( v, 0xffff);
}

public static void expandPath ( Vector v, LeveledObject lo) {
	if ( (v == null) || (lo == null) )
		return;

	int vs = v.size();
	int ll = lo.Level;

	if ( ll == 0)
		return;

	for ( int idx=v.indexOf( lo); idx>-1; idx--) {
		LeveledObject loc = (LeveledObject)v.elementAt( idx);
		if ( loc.Level == ll-1) {
			expandToLevel( v, vs, idx, loc.Level, ll);
			if ( --ll == 0)
				return;
		}
	}
}

public static void expandToLevel (Vector v, LeveledObject sObj, int maxLvl) {
	if ( (v != null) && (sObj != null) )
		expandToLevel( v, v.size(), v.indexOf( sObj), sObj.Level, maxLvl);
}

public static void expandToLevel (Vector v, int lvl) {
	int s = (v == null) ? 0 : v.size();

	for ( int i=0; i<s; i++){
		LeveledObject lo = (LeveledObject)v.elementAt( i);
		if (lo.Level <= lvl)
			lo.setVisible( true);
		if ((lo.Level < lvl) && lo.isExpandable() )
			lo.setExpanded( true);
	}
}

public static void expandToLevel (Vector v, int vs, int sIdx, int sLvl, int maxLvl) {
	for ( int idx=sIdx; idx<vs; idx++){
		LeveledObject lo = (LeveledObject)v.elementAt( idx);
		if ( (idx != sIdx) && (lo.Level <= sLvl) )
			break;
		if (lo.Level <= maxLvl)
			lo.setVisible( true);
		if ((lo.Level < maxLvl) && (lo.State == (LO_VISIBLE | LO_EXPANDABLE)) )
			lo.setExpanded( true);
	}
}

public static int getMaxLevel ( Vector v ) {
	int maxLvl = 0;
	int s = (v == null) ? 0 : v.size();
	for ( int i=0; i<s; i++) {
		LeveledObject lo = (LeveledObject)v.elementAt( i);
		if ( lo.Level > maxLvl)
			maxLvl = lo.Level;
	}
	return maxLvl;
}

public Object getObject () {
	return Obj;
}

public static void initialize (Vector v) {
	int s = v.size();

	for ( int i=0; i<s; i++){
		LeveledObject lo = (LeveledObject)v.elementAt( i);
		if ( lo.Level == 0)
			lo.setVisible( true);
		if ( i<s-1){
			LeveledObject lon = (LeveledObject)v.elementAt( i+1);
			if ( lon.Level > lo.Level){
				lon.Level = lo.Level + 1;
				lo.setExpandable( true);
			}
		}
	}  
}

public static void insert (Vector v, Object obj, int lvl, int idx) {
	LeveledObject ilo;
	int iLvl = 0;

	try {
		ilo = (LeveledObject)v.elementAt( idx-1);
		iLvl = Math.min( ilo.Level+1, lvl);
	}
	catch ( Throwable t) {}

	LeveledObject nlo = new LeveledObject( obj, iLvl);
	int ll = iLvl;

	if ( iLvl == 0)
		nlo.setVisible( true);

	for ( int i=idx-1; i>-1; i--){
		LeveledObject lo = (LeveledObject)v.elementAt( i);
		if ( lo.Level >= ll)
			break;
		lo.setExpandable( true);
		ll = lo.Level;
	}

	v.insertElementAt( nlo, idx);
}

public boolean isExpandable () {
	return ( State == (LO_VISIBLE | LO_EXPANDABLE));
}

public boolean isExpanded () {
	return ( (State & LO_EXPANDED) != 0);
}

public boolean isShrinkable () {
	return ( State == (LO_VISIBLE | LO_EXPANDABLE | LO_EXPANDED));
}

public boolean isVisible () {
	return ( (State & LO_VISIBLE) != 0);
}

public static boolean lastLevelIdx (Vector v, int idx, int lvl) {
	int s = v.size();

	for ( int i=idx+1; i<s; i++){
		LeveledObject lo = (LeveledObject)v.elementAt( i);
		if ( lo.Level < lvl)     return true;
		if ( lo.Level == lvl)    return false;
	}

	return true;
}

public int level () {
	return Level;
}

static void reset (Vector v) {
	int s = v.size();

	for ( int i=0; i<s; i++){
		LeveledObject lo = (LeveledObject)v.elementAt( i);
		lo.subState( LO_VISIBLE | LO_EXPANDED);
	}
}

public void setExpandable () {
	State |= LO_EXPANDABLE;
}

public void setExpandable ( boolean state) {
	if ( state)  State |= LO_EXPANDABLE;
	else         State &= ~LO_EXPANDABLE;
}

public void setExpanded ( boolean state) {
	if ( state)  State |= LO_EXPANDED;
	else         State &= ~LO_EXPANDED;
}

public void setVisible ( boolean state) {
	if ( state)  State |= LO_VISIBLE;
	else         State &= ~LO_VISIBLE;
}

public static void shrinkAll ( Vector v ) {
	reset( v);
	expandToLevel( v, 0);
}

public void subState (int flags) {
	State &= ~flags;
}
}
