package biss.awt;

import biss.ObserverSocket;
import biss.VectorLib;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * replaces java.awt.List. Far more general because it can deal with
 * Vectors of arbitrary objects. All processing (drawing, selection,
 * processing, hotkey commands, popup menus etc.) is done via
 * ObserverSockets. In case of Vector items implementing the SelfDrawing
 * interface (or of Strings, of course), nothing has to be done to get
 * the items displayed. The List also does not make a distinction
 * between single or multiple select mode. Instead of this, a selection
 * will be "remembered" (marked) if the <ctrl> key was pressed in combination
 * with the mouse button. <esc> resets all selections, <F12> selects all
 * items.
 *
 * (C) 1996,97 BISS GmbH Germany, see file 'LICENSE.BISS-AWT' for details
 * @author J.H.Mehlitz
 */
public class List
  extends ScrollablePane
{
	protected ListCanvas Cv;
	public Color DrawClrBack;
	public Color DrawClrFore;
	public Object DrawObject;
/**
 * triggered by an item paint request, just in case
 * of at least one registered observer.
 * This observer is responsible for all item painting
 * ( overrides the standard string representation)
 * parameter: list item
 */
	public ObserverSocket OsDrawObject = new ObserverSocket( this);
/**
 * triggered by an ACTION event (mouse1DblClick or Enter key).
 * parameter: selection
 */
	public ObserverSocket OsProcess = new ObserverSocket( this);
/**
 * triggered by a SELECT / DESELECT event (mouse1Up). Note that cleaning
 * the selection (e.g. ESC key) triggers the Socket with a null
 * parameter.
 * parameter: selection or null ( selection cleared)
 */
	public ObserverSocket OsSelection = new ObserverSocket( this);
	public boolean DrawingSelected = false;
	public boolean DrawingMarked = false;
/**
 * triggered by a mouse SELECT event( mouse1Up)
 * parameter: mouse selection point
 */
	public ObserverSocket OsMouseSelection = new ObserverSocket( this);
	public int DrawObjectIdx;
	public ListHeader Header;

public List(){
	this( SB_BOTH, true, null);
}

public List( String items){
	this( SB_BOTH, true, null);

	Vector v = new Vector();

	StringTokenizer tok = new StringTokenizer( items, ",:;");
	while( tok.hasMoreElements() )
		v.addElement( tok.nextToken() );
	setContents( v);
}

public List( int type) {
	this( type, true, null);
}

public List( int type, Title t) {
	this( type, true, t);
}

protected List( int type, boolean setCanvas) {
	this ( type, setCanvas, null);
}

protected List( int type, boolean setCanvas, Title t) {
	super( type);

	setTitle( t);
	if ( setCanvas) {
		Cv = new ListCanvas( Horz, Vert, this);
		setCanvas( Cv);
	}	
}

public void addAllElements ( Vector v ) {
	if ( Cv.Lines == null){
		Cv.Lines = new Vector();
		Cv.setLineDimensions( v.firstElement() );
	}
	VectorLib.addAllElements( Cv.Lines, v);
	Cv.updateContents();
}

public void addElement ( Object obj ) {
	if ( Cv.Lines == null) {
		Cv.Lines = new Vector();
		Cv.setLineDimensions( obj);
	}
	Cv.Lines.addElement( obj);
	Cv.updateVScroll();
	if ( Cv.isShowing() )
		Cv.redrawObject( obj);
}

public void clearMarked () {
	Cv.clearMarked();
}

public boolean clearMarked ( Object obj ) {
	return Cv.clearMarked( obj);
}

public void clearSelections ( boolean notify ) {
	Cv.clearMarked();
	setSelection( (Object)null, notify);
}

public void deselect ( int index) {
	Cv.deselect( index);
}

public Object elementAt( int index) {
	if ( Cv.Lines == null)
		return null;
	return Cv.Lines.elementAt( index);
}

public Vector getContents() {
	return Cv.Lines;
}

public Rectangle getDrawRectFor ( Object o ) {
	return Cv.drawRectFor( o);
}

public int getDrawWidth () {
	return Cv.Width;
}

public Object getLastSelection() {
	return Cv.LastSelection;
}

public int getLastSelectionIndex() {
	if ( Cv.LastSelection != null)
		return Cv.Lines.indexOf( Cv.LastSelection);
	return -1;
}

public Dimension getMinimumSize () {
	return getMinimumSize( 5);
}

public Dimension getMinimumSize (int v) {
	Font fnt = Cv.getFont();
	FontMetrics fm = Awt.getFontMetrics( fnt != null ? fnt : Awt.SysFont );
	int lh = Cv.LineHeight != 0 ? Cv.LineHeight : fm.getHeight();
	int cx = 15 + fm.charWidth( 'x') * 15;
	int cy = 15 + lh * v;
	
	return new Dimension( cx, cy);
}

public Dimension getPreferredSize( int v){
	return getMinimumSize( v);
}

public int[] getSelectedIndexes () {
	Vector v = getSelections();
	int[]  ia = null;
	if ( v != null){
		ia = new int[v.size()];
		for ( int i=0; i<v.size(); i++)
			ia[i] = Cv.Lines.indexOf( v.elementAt( i));
	}
	return ia;
}

public Object getSelection(){
	return Cv.Selection;
}

public int getSelectionIndex() {
	if ( Cv.Selection != null)
		return Cv.Lines.indexOf( Cv.Selection);
	return -1;
}

public Vector getSelections () {
	return Cv.Selections;
}

public boolean hasSelection () {
	return (Cv.Selection != null);
}

public boolean hasSelections () {
	return !Cv.Selections.isEmpty();
}

public void insertAllElementsBefore ( Vector v, Object nextObj ) {
	if ( Cv.Lines == null)
		Cv.Lines = new Vector();
	VectorLib.insertAllElementsBefore( Cv.Lines, v, nextObj);
	Cv.updateContents();
}

public void insertAllElementsBehind ( Vector v, Object obj ) {
	if ( Cv.Lines == null)
		Cv.Lines = new Vector();
	VectorLib.insertAllElementsBehind( Cv.Lines, v, obj);
	Cv.updateContents();
}

public void insertElementBefore ( Object insObj, Object prevObj ) {
	int insIdx = (Cv.Lines != null) ? Cv.Lines.indexOf( prevObj) : 0; 
	insertElementBefore( insObj, insIdx);
}

public void insertElementBefore ( Object insObj, int insIdx ) {
	if ( Cv.Lines == null)
		Cv.Lines = new Vector();
	if ( insIdx < 0)
		insIdx = 0;
	Cv.Lines.insertElementAt( insObj, insIdx);
	Cv.updateContents();
}

public boolean isIndexSelected (int index) {
	try { return Cv.Selections.contains( Cv.Lines.elementAt( index)); }
	catch( Exception e) { return false; }	
}

public void makeVisible (int index) {
	showObject( elementAt( index));
}

public boolean mouseSelection(){
	return (Cv.SelSource == ListCanvas.OP_SEL_MOUSE);
}

void notifyProcess () {
	if ( Cv.Selection != null) {
		Object arg;
		if ( Cv.Selection instanceof SelfDrawingObject)
			arg = ((SelfDrawingObject) Cv.Selection).label();
		else
			arg = Cv.Selection.toString();
		OsProcess.notifyObservers( Cv.Selection);
	}
}

void notifySelect () {
	Object arg = null;

	if ( Cv.Selection != null) {
		if ( Cv.Selection instanceof SelfDrawingObject)
			arg = ((SelfDrawingObject) Cv.Selection).label();
		else
			arg = Cv.Selection.toString();
	}

	OsSelection.notifyObservers( Cv.Selection);
}

public int numberOfDisplayedItems () {
	return Cv.Height / Cv.LineHeight;
}

public void overrideDrawSocket ( boolean state) {
	Cv.OverrideDrawSocket = state;
}

public int position() {
	return Cv.IdxTop;
}

public Dimension preferredSize (int v) {
	int lh = Cv.LineHeight != 0 ? Cv.LineHeight : Awt.SysFontHeight;
	return new Dimension( 10*Awt.SysFontWidth, lh*v);
}

public void redrawObject ( Object obj ){
	Cv.redrawObject( obj);
}

public void removeAllElements () {
	Cv.updateContents( (Vector)null);
}

public void removeAllElements ( Vector list ) {
	if ( list == null)
		return;
	if ( Cv.Lines != null)
		VectorLib.removeAllElements( Cv.Lines, list);
	Cv.updateContents();
}

public void removeElement ( Object obj ) {
	if ( Cv.Lines != null) {
		Cv.Lines.removeElement( obj);
		Cv.updateContents();
	}
}

public void removeElementAt ( int index ) {
	if ( Cv.Lines != null) {
		Cv.Lines.removeElementAt( index);
		Cv.updateContents();
	}
}

public void removeElements ( int start, int end ) {
	if ( Cv.Lines != null) {
		for ( int i=start; i<=end; i++)
			Cv.Lines.removeElementAt( i);
		Cv.updateContents();
	}
}

public void replaceElementAt ( Object newObj, int index ) {
	int insIdx;
	if ( (Cv.Lines == null) || (index < 0) )
		return;
	Cv.Lines.setElementAt( newObj, index);
	if ( index == getSelectionIndex() )
		Cv.Selection = newObj;
	Cv.redrawObject( newObj);
}

public void replaceElementWith ( Object oldObj, Object newObj ) {
	int insIdx = (Cv.Lines != null) ? Cv.Lines.indexOf( oldObj) : -1;
	replaceElementAt( newObj, insIdx); 
}

public void reset () {
	//clear selections and position home
	Cv.reset();
}

public void restoreLastSelection ( boolean fireSlot ) {
	setSelection( Cv.LastSelection, fireSlot);
}

public void scrollVertical ( int nLines ) {
	Cv.scrollVertical( nLines, true);
}

public void select (int index) {
	setSelection( elementAt( index), true);
}

public void selectAll ( boolean notify ) {
	Cv.selectAll( notify);
}

public synchronized void setBackground ( Color c ) {
	Cv.setBackground( c);
}

public void setColors( Color cFore, Color cBack, Color cMFore,
		Color cMBack, Color cSFore, Color cSBack) {
	Cv.setColors( cFore, cBack, cMFore, cMBack, cSFore, cSBack);
}

public int setContents( Vector data){
	return Cv.setContents( data);
}

public boolean setDrawEnvironment ( Object obj ) {
	return Cv.setDrawEnvironment( obj);
}

public void setDrawHeight ( int height ) {
	Cv.setLineHeight( height);
}

public void setDrawWidth ( int width ) {
	Cv.LineWidth = width;
}

public synchronized void setForeground ( Color c ) {
	Cv.setForeground( c);
}

public boolean setMarked ( Object obj ) {
	return Cv.setMarked( obj);
}

public boolean setSelection ( Object obj, boolean fireSlot ) {
	return Cv.setSelection( obj, fireSlot);
}

public boolean setSelection( int idx, boolean fireSlot){
	return Cv.setSelection( idx, fireSlot);
}

public boolean shiftDown () {
	int  selIdx;

	if ( Cv.Selection == null)
		return false;

	selIdx = Cv.Lines.indexOf( Cv.Selection);
	if ( selIdx < Cv.Lines.size()-1) {
		VectorLib.swap( Cv.Lines, selIdx, selIdx+1);
		Cv.updateContents();
		Cv.showArea( selIdx, selIdx+1);
		return true;
	}

	return false;
}

public boolean shiftUp () {
	int  selIdx;

	if ( (selIdx = Cv.Lines.indexOf( Cv.Selection)) > 0) {
		VectorLib.swap( Cv.Lines, selIdx-1, selIdx);
		Cv.updateContents();
		Cv.showArea( selIdx-1, selIdx);
		return true;
	}

	return false;
}

public boolean showIndex ( int index ) {
	try {
		Cv.showObject( Cv.Lines.elementAt( index) );
		return true;
	}
	catch ( Exception e) { return false; }
}

public void showLastObject () {
	Cv.showLastObject();
}

public boolean showObject ( Object obj ) {
	return Cv.showObject( obj);
}

public boolean toggleMarker ( Object obj ) {
	return Cv.toggleMarker( obj);
}

public void updateContents () {
	Cv.updateContents();
}

public void updateContents ( Vector newList) {
	Cv.updateContents( newList);
}
}

class ListCanvas
  extends ScrollableCanvas
  implements HotKeyProcessor
{
	Color ColorBackMark = Color.lightGray;
	Color ColorBackSel = Awt.SelBackClr;
	Color ColorForeMark = Color.black;
	Color ColorForeSel = Awt.SelForeClr;
	List Master;
	StringBuffer FLNBuf = new StringBuffer();
	Object LastSelection;
	final public static int OP_SEL_KBD = 1;
	final public static int OP_SEL_MOUSE = 2;
	public int SelSource;
	Object Selection;
	Vector Selections = new Vector();
	Rectangle WorkRect = new Rectangle();
	boolean OverrideDrawSocket = false;
	boolean LastIsMarked = true;
	Font DefDrawFont;
	int DefDrawHeight;
	int DefDrawDescent;

ListCanvas( HScrollBar h, VScrollBar v, List cmpd) {
	super(h, v);
	setForeground( Color.black);
	setBackground( Color.white);
	Master = cmpd;
	
	registerHotKeys();
}

public void charInput( char c, int mod) {
	/* first letter navigation
	 * type lowercase char for initial position
	 * optionally followed by uppercase chars for extended search
	*/
	boolean uc = ( (c & 32) == 0);
	char lc = (char)( c | 32);

	if ( ! uc)
		FLNBuf.setLength( 0);
	
	FLNBuf.append( lc);
	findFLN();
}

boolean clearMarked(){
	if ( Selections.size() < 2)
		return false;
		
	Selections.removeAllElements();
	if ( Selection != null)
		Selections.addElement( Selection);

	redraw();
	return true;
}

boolean clearMarked( Object obj) {
	if ( obj == null)
		return false;

	int s = Selections.size();
	if ( !LastIsMarked) s--;
	for ( int i=0; i<s; i++) {
		if ( obj == Selections.elementAt( i)){
			Selections.removeElement( obj);
			redrawObject( obj);
			return true;
		}
	}
	
	return false;
}

protected void clearSelection() {
	LastSelection = Selection;
	Selection = null;

	if ( LastSelection != null){
		if ( !LastIsMarked) {
			Selections.removeElementAt( Selections.size()-1);
			LastIsMarked = true;
		}
		drawObject( LastSelection);
	}
}

void clearSelections( boolean redraw ){

	int s = Selections.size();

	Selections.removeAllElements();
	LastIsMarked  = true;
	LastSelection = Selection;
	Selection     = null;

	if ( redraw) {
		if ( s>1) redraw();
		else if ( LastSelection != null) redrawObject( LastSelection);
	}
}

void deselect (int idx){
	if ( Lines == null)
		return;

	Object obj = Lines.elementAt( idx);
	if ( obj == Selection)
		clearSelection();
	else
		clearMarked( obj);
}

synchronized boolean drawObject( Object obj) {
	if ( !setDrawEnvironment( obj) )
		return false;
		
	drawRenderedObject();
	return true;
}

Rectangle drawRectFor( Object obj) {
	int idx = objectIndexOfVisible( obj);
	
	if ( idx >= 0 )
		return drawRectForIndex( idx);
	else
		return null;
}

Rectangle drawRectForIndex ( int idx ) {

	WorkRect.x = XOffset;
	WorkRect.y = DrawRect.y + (idx - IdxTop) * LineHeight;
	WorkRect.width  = Width - XOffset - FBorder.Ext;
	WorkRect.height = LineHeight;
	
	return WorkRect;
}

void drawRenderedObject () {
	Object    obj = Master.DrawObject;
	Rectangle dRect = Master.DrawRect;

	ResGraph.setColor( Master.DrawClrBack);
	ResGraph.fillRect( dRect.x, dRect.y, dRect.width, dRect.height);
	ResGraph.setColor( Master.DrawClrFore);

	if ( (! OverrideDrawSocket) && (! Master.OsDrawObject.isEmpty()) )
		Master.OsDrawObject.notifyObservers( Master);
	else if ( obj instanceof SelfDrawingObject)
		((SelfDrawingObject) obj).drawSelfIn( Master);
	else {
		ResGraph.setFont( DefDrawFont);
		ResGraph.drawString( obj.toString(), dRect.x + 5,
		                     dRect.y + dRect.height -
		                     (dRect.height - DefDrawHeight) / 2 -
		                     DefDrawDescent );
	}
}

protected boolean findFLN(){
	return selectLabel( FLNBuf.toString());
}

boolean hasStringList(){
	if ( (Lines == null) || (Lines.size() == 0) )
		return false;
	return ( Lines.elementAt( 0) instanceof String);
}

boolean isMarked (Object obj){
	if ( obj == null)
		return false;
		
	int s = Selections.size();
	if ( ! LastIsMarked) s--;
		
	for ( int i=0; i<s; i++){
		if ( obj == Selections.elementAt( i))
			return true;
	}
	
	return false;
}

public void mouse1DblClick() {
	Master.notifyProcess();
}

public void mouse1Down() {
	Object ns = objectAt( Mouse.Pos.x, Mouse.Pos.y);
	SelSource = OP_SEL_MOUSE;

	if ( Key.isCtrlDown() ){
		toggleMarker( ns);
		Master.notifySelect();
	}
	else
		selectObject( ns, false);
}

public void mouse1Up() {
	int x = Mouse.Pos.x;
	int y = Mouse.Pos.y;
	int idx = objectIndexAt( x, y);
	Object ns = (idx < 0) ? null : Lines.elementAt( idx);
	Rectangle r;
	Point pt;

	if ( ns == Selection)
		Master.notifySelect();

	if ( ns != null){
		r = drawRectForIndex( idx);
		if ( r != null ) {
			pt = new Point( x - r.x, y - r.y);
			Master.OsMouseSelection.notifyObservers( pt);
		}
	}
}

public void processKey ( Object sel) {
	if ( sel.equals( "toggle"))
		toggleMarker( Selection);
	else if ( sel.equals( "process")){
		if (Selection != null)
			Master.notifyProcess();
	}
	else if ( sel.equals( "escape"))
		reset();
	else if ( sel.equals( "selectAll"))
		selectAll( true);
	else if ( sel.equals( "home"))
		home();
	else if ( sel.equals( "end"))
		end();
	else if ( sel.equals( "pageUp"))
		pageUp();
	else if ( sel.equals( "pageDown"))
		pageDown();
	else if ( sel.equals( "cursorUp"))
		shiftSelection( -1);
	else if ( sel.equals( "cursorDown"))
		shiftSelection( 1);
	else if ( sel.equals( "cursorLeft")){
		if ( Horz != null)
			Horz.pageDown();
	}
	else if ( sel.equals( "cursorRight")){
		if ( Horz != null)
			Horz.pageUp();
	}
	else
		super.processKey( sel);
}

void redrawLine ( int idx ){
	if ( setDrawEnvironment( idx, drawRectForIndex( idx)) )
		drawRenderedObject();
}

void redrawObject ( Object obj ) {
	drawObject( obj);
}

void registerHotKeys () {
	EventHandler.addHotKey( Awt.ListSelToggleKeys[0], Awt.ListSelToggleKeys[1], this, "toggle");
	EventHandler.addHotKey( Awt.ListProcessKeys[0], Awt.ListProcessKeys[1], this, "process");
	EventHandler.addHotKey( Awt.ListSelClearKeys[0], Awt.ListSelClearKeys[1], this, "escape");
	EventHandler.addHotKey( Awt.ListSelAllKeys[0], Awt.ListSelAllKeys[1], this, "selectAll");
	
	EventHandler.addHotKey( Key.Home, HotKey.Lone, this, "home");
	EventHandler.addHotKey( Key.End, HotKey.Lone, this, "end");
	EventHandler.addHotKey( Key.PageUp, HotKey.Lone, this, "pageUp");
	EventHandler.addHotKey( Key.PageDown, HotKey.Lone, this, "pageDown");
	EventHandler.addHotKey( Key.CursorUp, HotKey.Lone, this, "cursorUp");
	EventHandler.addHotKey( Key.CursorDown, HotKey.Lone, this, "cursorDown");
	EventHandler.addHotKey( Key.CursorLeft, HotKey.Lone, this, "cursorLeft");
	EventHandler.addHotKey( Key.CursorRight, HotKey.Lone, this, "cursorRight");
}

void reset () {
	clearSelections( true);
	positionHorizontal( 0, true);
	positionVertical( 0, true);
	Master.notifySelect();
}

boolean selectAll ( boolean notify ) {
	if ( Lines == null)
		return false;
		
	int s = Lines.size();
	int si = (Selection != null) ? Lines.indexOf( Selection) : -1;
	
	Selections.removeAllElements();
	LastIsMarked = true;
	
	for ( int i=0; i<s; i++) {
		if ( si != i)
			Selections.addElement( Lines.elementAt( i));
	}
	if ( si>-1){
		Selections.addElement( Selection);
		LastIsMarked = false;
	}

	redraw();

	if ( notify )
		Master.notifySelect();
		
	return true;
}

boolean selectLabel( String label) {
	Object obj;
	SelfDrawingObject sdo;
	String lStr;

	if ( (label == null) || (Lines == null) || (Lines.size() == 0))
		return false;

	label = label.toLowerCase();

	for ( Enumeration e = Lines.elements(); e.hasMoreElements(); ) {
		obj = e.nextElement();

		if ( obj instanceof SelfDrawingObject)
			lStr = ((SelfDrawingObject) obj).label();
		else
			lStr = obj.toString();

		if ( lStr.toLowerCase().startsWith( label) ){
			SelSource = OP_SEL_KBD;
			if ( obj != Selection)
				setSelection( Lines.indexOf( obj), true);
			return true;
		}
	}

	return false;						
}

protected void selectObject( Object obj, boolean notify ) {

	if ( obj == Selection) {
		LastSelection = Selection;
		return;
	}
		
	clearSelection();
	Selection = obj;

	if ( Selection != null){
		if ( ! isMarked( Selection)){
			Selections.addElement( Selection);
			LastIsMarked = false;
		}
		drawObject( Selection);
	}

	if ( notify )
		Master.notifySelect();
}

void setColors( Color cf, Color cb, Color cmf, Color cmb, Color csf, Color csb) {

	ColorForeMark = cmf;
	ColorBackMark = cmb;
	ColorForeSel  = csf;
	ColorBackSel  = csb;

	setBackground( cb);
	setForeground( cf);
}

synchronized int setContents( Vector data ) {
	clearSelections( false);

	Lines   = data;
	IdxTop  = 0;
	XOffset = XBorder + FBorder.Ext;

	if ( (Lines == null) || (Lines.size() == 0)){
		setLineDimensions( null);
		if (Vert != null)
			Vert.setValues( 0, 0, 0);
		if ( ResGraph != null)
			redraw();
		return 0;
	}

	setLineDimensions( Lines.firstElement() );

	if ( ResGraph != null )
		redraw();

	return Lines.size();
}

boolean setDrawEnvironment ( Object obj ) {
	int idx = objectIndexOfVisible( obj);
	
	if ( idx >= 0 )
		return setDrawEnvironment( idx, drawRectForIndex( idx));

	return false;
}

boolean setDrawEnvironment ( int idx, Rectangle dRect ) {
	Object    obj;

	if ( dRect == null || ResGraph == null )
		return false;
	
	obj = Lines.elementAt( idx);

	Master.DrawingSelected = Master.DrawingMarked = false;

	if ( obj == Selection){
		Master.DrawClrFore = ColorForeSel;
		Master.DrawClrBack = ColorBackSel;
		Master.DrawingSelected = true;
	}
	else if ( Selections.contains( obj) ){
		Master.DrawClrFore = ColorForeMark;
		Master.DrawClrBack = ColorBackMark;
		Master.DrawingMarked = true;
	}
	else {
		Master.DrawClrFore = getForeground();
		Master.DrawClrBack = getBackground();
	}

	Master.DrawObject = obj;
	Master.DrawObjectIdx = idx;
	Master.DrawRect = dRect;
	Master.DrawGraphics = ResGraph;

	return true;
}

public synchronized void setFont ( Font f) {
	DefDrawFont = (f != null) ? f : Awt.SysFont;
	FontMetrics fm = Awt.DefToolkit.getFontMetrics( DefDrawFont);
	DefDrawHeight = fm.getHeight();
	DefDrawDescent = fm.getMaxDescent();
}

void setLineDimensions ( Object refObj ) {
	if ( refObj != null && refObj instanceof SelfDrawingObject ){
		SelfDrawingObject fo = (SelfDrawingObject) refObj;
		if ( LineHeight == 0)
			LineHeight = fo.drawHeight();
		if ( LineWidth == 0)
			LineWidth = fo.maxDrawWidth();    
	}
	else {
		if ( LineHeight == 0)
			LineHeight = Awt.SysFontHeight;
		if ( LineWidth == 0)
			LineWidth = Awt.ScreenWidth;
	}

	PageInc = Height / LineHeight;

	if ( Vert != null){
		if ( Lines != null)
			Vert.setValues( 0, 0, Lines.size() - PageInc);
		Vert.setStepPels( LineHeight);
		Vert.setPageIncrement( PageInc);
	}
	if ( Horz != null){
		Horz.setValues( 0, 0, LineWidth);
		Horz.setLineIncrement( 10);
		Horz.setPageIncrement( 100);
	}

}

boolean setMarked( Object obj) {
	if ( (obj == null) || isMarked( obj) )
		return false;

	if ( LastIsMarked)
		Selections.addElement( obj);
	else if ( obj != Selection)
			Selections.insertElementAt( obj, Selections.size()-1 );
	else
		LastIsMarked = true;
		
	redrawObject( obj);
	return true;
}

boolean setSelection( Object obj, boolean fire) {
	if ( Lines == null ){
		if ( obj == null ){
			if ( fire)
				Master.notifySelect();
			return true;
		}
		return false;
	}

	if ( obj == null ) {
		clearSelections( true);
		if ( fire)
			Master.notifySelect();
		return true;
	}

	return setSelection( Lines.indexOf( obj), fire);
}

boolean setSelection( int idx, boolean fire) {

	if ( Lines == null)
		return false;

	if ( (idx < 0) || (idx >= Lines.size()) )
		return false;

	showIndex( idx);
	selectObject( Lines.elementAt( idx), fire);

	return true;
}

boolean shiftSelection( int steps) {

	if ( Lines == null)
		return false;

	SelSource = OP_SEL_KBD;

	if ( Selection == null)
		setSelection( 0, true);
	else
		setSelection( Lines.indexOf( Selection) + steps, true);

	return true;
}

void showLastObject () {
	int n;

	if ( (Lines == null) || ( (n = Lines.size()) == 0))
		return;
	showArea( n, n);
}

boolean showObject ( Object obj ) {
	if ( (Lines == null) || ( ! Lines.contains( obj)) )
		return false;

	showIndex( Lines.indexOf( obj) );
	return true;
}

boolean toggleMarker( Object obj) {
	if ( obj == null)
		return false;

	if ( isMarked( obj) )
		clearMarked( obj);
	else
		setMarked( obj);

	return true;
}

void updateContents () {
	updateContents( false);
}

void updateContents ( Vector newList) {
	Lines = newList;
	if ( Lines == null)
		setContents( null);
	else
		updateContents( false);
}

void updateContents ( boolean notifyChange) {
	boolean chg = false;

	if ( LineHeight == 0 ){  // first time
		setContents( Lines);
		return;
	}
	
	for ( int i=0; i<Selections.size(); i++){
		Object mrk = Selections.elementAt( i);
		if ( (Lines == null) || (! Lines.contains( mrk)) ){
			if ( mrk == Selection){
				LastIsMarked = true;
				LastSelection = null;
				Selection = null;
			}
			Selections.removeElementAt( i--);
			chg = true;
		}
	}

	if ( Lines == null)
		IdxTop = 0;
	else if ( Lines.size() < IdxTop)
		IdxTop = Lines.size();

	updateVScroll();
	redrawLines( IdxTop, -1); 
	
	if ( chg && notifyChange)
		Master.notifySelect();
}

void updateScrollImage( int sIdx){
	Graphics g = ResGraph;
	ResGraph = SoftGraph;

	int y0 = 0;
	for ( int i=0; i<2; i++) {
		WorkRect.x = XOffset;
		WorkRect.y = y0;
		WorkRect.width  = Width;
		WorkRect.height = LineHeight;

		try {
			if ( setDrawEnvironment( sIdx+i, WorkRect) )
				drawRenderedObject();
		}
		catch ( Throwable t) {
			ResGraph.setColor( getBackground() );
			ResGraph.fillRect( WorkRect.x, WorkRect.y,
			                   WorkRect.width, WorkRect.height);
		}
		y0 += LineHeight;
	}

	ResGraph = g;
}
}
