package biss.awt;

import biss.ObserverSocket;
import java.awt.Cursor;
import java.awt.Event;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Frame;
import java.awt.Graphics;
import java.util.StringTokenizer;
import java.util.Vector;

class ListColumn
{
	public int XPos;
	public int Width;
	public String Label;
	public boolean Checked;
	public int MinWidth;

public ListColumn ( String label, int x, int w) {
	this( label, x, w, w);
}

public ListColumn ( String label, int x, int w, int mw) {
	Label = label;
	XPos = x;
	MinWidth = mw;
	Width = w;
}

void draw ( Graphics g, int offs, int height, Font fnt, boolean carved) {
	int x = XPos + offs;
	GraphicsLib.paint3dBox( g, x, offs, Width, height, Checked, Awt.ButtonBackClr);
	GraphicsLib.drawStringCentered( g, Label, x, 0, Width, height, fnt, Awt.ButtonForeClr, carved);
}
}

/**
 * ListHeader widget ( used e.g. in ColumnList), for 'columnized'
 * views of a List. Columns can be sized at runtime ( mouse1Draw )
 *
 * (C) 1996,97 BISS GmbH Germany, see file 'LICENSE.BISS-AWT' for details
 * @author J.H.Mehlitz
 */
public class ListHeader
  extends Widget
  implements MouseMotionInterface
{
	Font LabelFnt = Awt.MenuFont;
	public ListColumn[] Columns;
	int SizeIdx = -1;
	public ObserverSocket OsWidthChange = new ObserverSocket( this);
	public ObserverSocket OsCheck = new ObserverSocket( this);
	boolean Checkable;

public ListHeader () {
	FBorder.setBorder( Border.BS_INSET);
	setForeground( Awt.ButtonForeClr);
	setBackground( Awt.ButtonBackClr);
}

public ListHeader ( String labels) {
	this();
	addColumns( labels);
}

public void addColumn ( String label) {
	FontMetrics fm = Awt.getFontMetrics( LabelFnt);
	addColumn( label, fm.stringWidth( label) + 3*fm.charWidth( ' '));
}

public void addColumn ( String label, int width) {
	
	if ( Columns == null){
		Columns = new ListColumn[1];
		Columns[0] = new ListColumn( label, 0, width);
	}
	else {
		ListColumn pc[] = Columns;
		ListColumn lc = null;
		Columns = new ListColumn[pc.length+1];
		for ( int i=0; i<pc.length; i++)
			lc = Columns[i] = pc[i];
		Columns[pc.length] = new ListColumn( label, lc.XPos + lc.Width, width);
	}
	
	if ( isShowing() )
		redraw();
}

public void addColumns ( String labels) {
	StringTokenizer tok = new StringTokenizer( labels, ",:;");
	while( tok.hasMoreElements() )
		addColumn( tok.nextToken() );
}

public void addColumns ( Vector v) {
	for ( int i=0; i<v.size(); i++) {
		Object o = v.elementAt( i);
		if ( o instanceof String)
			addColumn( (String) o );
	}
}

boolean checkSizePos ( int x) {
	if ( Columns == null)
		return false;
		
	int dx = FBorder.Ext;
	for ( int i=0; i<Columns.length; i++) {
		ListColumn lc = Columns[i];
		int xr = lc.XPos + lc.Width;
		if ( (xr-dx<x) && (xr+dx>x) ) {
			SizeIdx = i;
			return true;
		}
	}
	
	return false;
}

ListColumn columnAtX ( int x) {
	if ( Columns == null)
		return null;
		
	for ( int i=0; i<Columns.length; i++) {
		ListColumn lc = Columns[i];
		if ( lc.XPos + lc.Width > x)
			return lc;
	}
	
	return null;
}

public int columnXPos ( int idx) {
	try { return Columns[idx].XPos; }
	catch ( Exception e) { return -1; }
}

public int countElements () {
	if ( Columns == null)
		return 0;
	return Columns.length;
}

public void mouse1Click () {
	if ( (SizeIdx == -1) && Checkable ){
		ListColumn lc = columnAtX (Mouse.Pos.x - FBorder.Ext);
		if ( lc != null) {
			lc.Checked = !lc.Checked;
			redraw();
			OsCheck.notifyObservers( lc);
		}
	}
}

public void mouse1Drag () {
	if ( SizeIdx < 0 )
		return;
		
	ListColumn lc = Columns[SizeIdx];
	int nw = Math.max( Mouse.Pos.x - FBorder.Ext - lc.XPos, lc.MinWidth);
	int dw = lc.Width - nw;

	if ( (nw == lc.MinWidth) && (lc.Width == nw) )
		return;
		
	lc.Width = nw;

	for ( int i=SizeIdx+1; i<Columns.length; i++)
		Columns[i].XPos -= dw;

	redraw();
}

public void mouse1Up () {
	if ( SizeIdx > -1 ){
		OsWidthChange.notifyObservers();
		SizeIdx = -1;
	}
}

public void mouse2Drag () {
}

public void mouseMove () {
	if ( checkSizePos (Mouse.Pos.x - FBorder.Ext))
		EventHandler.setCursor( Cursor.E_RESIZE_CURSOR);
	else
		EventHandler.setCursor( Cursor.DEFAULT_CURSOR);
}

public int preferredHeight () {
	return 3*Awt.SysFontHeight/2;
}

public void redraw ( Graphics g) {
	if ( Columns == null){
		blank( g);
		return;
	}

	ListColumn lc = null;
	clip( g);

	for ( int i=0; i<Columns.length; i++) {
		lc = Columns[i];
		if ( lc.XPos > DrawRect.width)
			return;
		if ( lc.XPos + lc.Width > 0)
				lc.draw( g, FBorder.Ext, DrawRect.height, LabelFnt, true );
	}

	g.setColor( getBackground() );
	g.fillRect( FBorder.Ext + lc.XPos + lc.Width, 0, Width, Height);
}

public void setCheckable ( boolean c) {
	Checkable = c;
}

void setOffset ( int x) {
	if ( Columns != null){
		int x0 = -x;
		for ( int i=0; i<Columns.length; i++) {
			ListColumn lc = Columns[i];
			lc.XPos = x0;
			x0 += lc.Width;
		}
	}
	redraw();
}
}
