package biss.awt;

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.util.*;

/**
 * abstract base for scrollable Canvas widgets inside of Compounds
 *
 * (C) 1996,97 BISS GmbH Germany, see file 'LICENSE.BISS-AWT' for details
 * @author J.H.Mehlitz
 */
abstract public class ScrollableCanvas
  extends Widget
{
	HScrollBar Horz;
	int IdxTop;
	int LineHeight;
	int LineWidth;
	Vector Lines;
	VScrollBar Vert;
	int XBorder;
	int XOffset;
	Graphics ResGraph;
	static Image SoftImage;
	static Graphics SoftGraph;
	static Dimension SoftSize;
	int PageInc;
/**
 * OS/2 workaround ( copyArea within Window fails )
 */
	boolean UseCopyArea = true;

public ScrollableCanvas( HScrollBar h, VScrollBar v) {
	super();
	Horz = h;
	Vert = v;

	setForeground( Color.black);
	setBackground( Color.white);

	registerObserver();
}

boolean atEnd() {
	if ( Lines == null)
		return true;
	return ( PageInc >= Lines.size() - IdxTop);
}

void blankTrailing() {
	int y = -1;

	if ( Lines == null)
		y = 0;
	else if (Lines.size() <= IdxTop + PageInc)
		y = (Lines.size() - IdxTop) * LineHeight + DrawRect.x;

	if ( y > -1) {	
		ResGraph.setColor( getBackground());
		ResGraph.fillRect( 0, y, Width, Height);
		ResGraph.setColor( getForeground());
	}				
}

void checkSoftImage() {
	int w, h;

	if ( SoftImage == null ) {
		w = Math.max( (2*Awt.ScreenWidth / 3), Width);
		h = Math.max( (2*Awt.SysFontHeight), LineHeight);
		SoftSize = new Dimension( w, h);
		SoftImage    = createImage( w, h);
		if ( SoftImage != null) 
			SoftGraph = SoftImage.getGraphics();
	}
	else {
		if ( SoftSize.width < Width || SoftSize.height < LineHeight ) {
			SoftSize.width  = Math.max( SoftSize.width, Width);
			SoftSize.height = Math.max( SoftSize.height, LineHeight);
			SoftImage    = createImage( SoftSize.width, SoftSize.height);
			SoftGraph = SoftImage.getGraphics();
		}
	}
}

void end() {
	if ( Lines != null){
		int idx = Lines.size() - 1;
		showArea( idx, idx);
	}
}

void home() {
	if ( IdxTop > 0)
		scrollVertical( -IdxTop, true); 			
}

boolean isItemVisible( int index) {
	if ( (Lines == null) || (LineHeight == 0) )
		return false;
	if ( index < IdxTop)
		return false;
	if ( index > IdxTop + PageInc)
		return false;
	return true;
}

void lineDown() {
	scrollVertical( 1, true); 			
}

void lineUp() {
	if ( IdxTop > 0)
		scrollVertical( -1, true);
}

Object objectAt( int x, int y) {
	int idx = objectIndexAt( x, y);
	
	if ( idx < 0 )
		return null;
	else
		return Lines.elementAt( idx);
}

int objectIndexAt( int x, int y) {

	if ( (x<0) || (x>Width) )
		return -1;
	if ( (y<0) || (y>Height) )
		return -1;
	if ( Lines == null)
		return -1;

	int offs = LineHeight + FBorder.Ext;

	for ( int idx = IdxTop; idx < Lines.size(); idx++) {
		if ( y <= offs)
			return idx;
		offs += LineHeight;
	}

	return -1;
}

int objectIndexOfVisible ( Object obj ) {
	int i;
	int ih = (Height - (FBorder.Ext<<1) + (LineHeight>>1)) / LineHeight;
	int n = Math.min( IdxTop + ih + 1, Lines.size());

	for ( i=IdxTop; i < n; i++ ) {
		if ( Lines.elementAt( i) == obj )
			return i;
	}
	
	return -1;
}

void pageDown() {
	scrollVertical( PageInc, true); 			
}

void pageUp() {
	if ( IdxTop > 0)
		scrollVertical( -PageInc, true);
}

public void paint( Graphics g){
	//first valid resident graphics can be obtained here
	if ( ResGraph == null) {
		updateGraphics();
		if ( Awt.SoftScroll )
			checkSoftImage();
	}

	super.paint( g);
}

protected void posChanged () {
	PageInc = (LineHeight > 0) ? Height / LineHeight : 0;	
	updateVScroll();

	//dispose resized (cliped) Graphics and obtain new one
	if ( ResGraph != null){
		ResGraph.dispose();
		updateGraphics();
	}
	if ( Awt.SoftScroll && (LineHeight > 0) )
		checkSoftImage();
}

boolean positionHorizontal( int pos, boolean updateScroll) {
	return scrollHorizontal( pos + XOffset - XBorder - FBorder.Ext, updateScroll);
}

boolean positionVertical( int pos, boolean updateScroll){
	return scrollVertical( pos - IdxTop, updateScroll);
}

public void redraw( Graphics g){
	if ( ResGraph == null)
		return;
	if (( LineHeight == 0) || (Lines == null) || (Lines.size() == 0))
		blank( ResGraph );
	else
		redrawLines( IdxTop, -1);
}

abstract void redrawLine( int idx);

synchronized void redrawLines( int start, int end) {

	int idx, max;

	if ( ResGraph == null )
		return;

	if ( Lines == null)
		max = -1;
	else	
		max = Math.min(IdxTop + PageInc, Lines.size() - 1);

	if ( end == -1)	end = max;
	else				end = Math.min( end, max);

	start = Math.max( start, IdxTop);

	for ( idx = start; idx <= end; idx++)
		redrawLine( idx);

	blankTrailing();
}

public void registerObserver(){
	if ( Vert != null)
		Vert.OsScroll.addObserver( this);
	if ( Horz != null)
		Horz.OsScroll.addObserver( this);

	super.registerObserver();
}

synchronized boolean scrollHorizontal( int steps, boolean updateScroll){
	int dx = XBorder + FBorder.Ext;

	if ( (steps == 0) || (Lines == null) )
		return false;

	if ( updateScroll && ( Horz != null) )
		Horz.updatePosition( steps - XOffset + dx);

	XOffset = Math.min( XOffset - steps, dx);
	redrawLines( IdxTop, -1);

	return true;
}

synchronized boolean scrollVertical( int steps, boolean updateScroll) {

	if ( (steps == 0) || (Lines == null))
		return false;

	int s = Lines.size();

	if ( steps < 0) {
		if ( IdxTop == 0)
			return false;
		if ( IdxTop + steps < 0)
			steps = -IdxTop;
	}
	else {
		if ( IdxTop + PageInc >= s )	//alredy all items visible
			return false;
		if ( IdxTop + steps >= s)
			steps = s - IdxTop - PageInc;
	}

	if ( updateScroll && (Vert != null) )
		Vert.updatePosition( IdxTop + steps);

	if ( ResGraph != null) {
		int as = Math.abs( steps);
		if ( (Vert != null) && Vert.IsSoftScroll )
			softScroll( steps);
		else if ( as <= PageInc ) {
			if ( UseCopyArea)	// *** OS/2 workaround ***
				ResGraph.copyArea( DrawRect.x, DrawRect.y, DrawRect.width, DrawRect.height,
			                   0, -steps * LineHeight);
		}
	}

	IdxTop += steps;
	vPosChange();

	if ( ! UseCopyArea)	// *** OS/2 workaround ***
		redrawLines( IdxTop, -1);
	else if ( steps < 0)
		redrawLines( IdxTop, IdxTop - steps - 1);
	else
		redrawLines( IdxTop + PageInc - steps - 1, -1);

	return true;
}

public void setBorder( int style ) {
	super.setBorder( style);
	XOffset = XBorder + FBorder.Ext;
}

void setLineHeight( int h) {
	LineHeight = h;

	PageInc = (LineHeight > 0) ? Height / LineHeight : 0;	
	updateVScroll();

	if ( (Lines != null) && (ResGraph != null ))
		redrawLines( IdxTop, -1);
}

public void setVisible( boolean b){
	super.setVisible( b);
	if ( ! b){
		if ( ResGraph != null){
			ResGraph.dispose();
			ResGraph = null;
		}
	}
}

void showArea( int sLine, int eLine) {
	if ( sLine < IdxTop)
		scrollVertical( sLine - IdxTop, true);
	else if ( eLine >= IdxTop + PageInc){
		int d = eLine-PageInc-IdxTop+1;
		scrollVertical( d, true);
	}
}

void showIndex( int index) {
	showArea( index, index);
}

void softScroll( int steps) {
	int imgY = 0;

	if ( steps == 1) {		//contents up
		updateScrollImage( IdxTop + PageInc);
		imgY = DrawRect.y + PageInc * LineHeight;
	}
	else if ( steps == -1) {	//contents down
		updateScrollImage( IdxTop - 2);
		imgY = DrawRect.y -2 * LineHeight ;
	}
	else
		return;

	for ( int i=1; i<=LineHeight; i++) {
		ResGraph.copyArea( DrawRect.x, DrawRect.y, DrawRect.width, DrawRect.height,
		                   0, -steps );
		imgY -= steps;
		ResGraph.drawImage( SoftImage, 0, imgY, this);

		// do an active defer (without explicitly rescheduling the current thread)
		// to avoid jerky scrolling
		if ( Vert != null) {
			Awt.sync();
			int n = Awt.SoftScrollDelay;
			for ( int k=0; k < n; k++ );
		}
	}

}

public void unregisterObserver(){
	if ( Vert != null)
		Vert.OsScroll.deleteObserver( this);
	if ( Horz != null)
		Horz.OsScroll.deleteObserver( this);

	super.unregisterObserver();
}

public void update( Observable o, Object arg){
	if ( (Vert != null) && (o == Vert.OsScroll) )
		scrollVertical( Vert.Pos - IdxTop, false);
	else if ( (Horz != null) && (o == Horz.OsScroll) )  
		scrollHorizontal( Horz.Pos + XOffset - XBorder - FBorder.Ext, false);
	else 
		super.update( o, arg);
}

void updateGraphics() {
		ResGraph = getGraphics();
		if ( ResGraph != null ){
			ResGraph.setFont( getFont() );
			clip( ResGraph);
		}
}

void updateScrollImage( int sIdx) {
}

void updateVScroll() {
	if ( (Vert != null) && (Lines != null) ){
		int oldMax = Vert.Max;
		
		Vert.setStepPels( LineHeight);
		Vert.setValues( IdxTop, PageInc, 0, Lines.size() - PageInc);
		
		if ( (oldMax != 0) && (Vert.Max == 0) )
			home();
	}
}

void vPosChange(){}
}
