package biss.awt;

import biss.ObserverSocket;
import java.awt.Color;
import java.awt.Container;
import java.awt.Event;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.LayoutManager;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.image.ImageObserver;
import java.util.Observable;
import java.util.Observer;
import java.util.Vector;

/**
 * abstract base for Compounds containing scrollable Canvas widgets
 *
 * There is an anomaly with respect to border handling. ScrollablePanes
 * don't have visible borders of their own. They just propagate their
 * border style to their ScrollableCanvas component.
 *
 * (C) 1996-1998 BISS GmbH Germany, see file 'LICENSE.BISS-AWT' for details
 * @author J.H.Mehlitz, M.Brauch
 */
abstract public class ScrollablePane
  extends Compound
  implements Observer, DrawablePane
{
	protected ScrollableCanvas BaseCv;
	protected int Type = SB_BOTH;
	protected HScrollBar Horz;
	protected VScrollBar Vert;
	protected Label Lab = new Label();
	protected Title InTitle = null;
/**
 * shadow of ScrollableCanvas OsCommand ( see Control)
 */
	public ObserverSocket OsCommand;
/**
 * triggered externally on demand by drawPane() method
 * parameter: -
 */
	public ObserverSocket OsDrawPane = new ObserverSocket( this);
	final public static int BarWidth = Awt.ScrollBarWidth;
	final public static int BarDist = 1;
	public static int DefaultBorderStyle = Border.BS_STEP;
	public Graphics DrawGraphics;
	public Rectangle DrawRect;
	final public static int SB_NONE = 3;
	final public static int SB_VERT = 1;
	final public static int SB_HORZ = 2;
	final public static int SB_BOTH = 0;

public ScrollablePane(){
	this( SB_BOTH);
}

public ScrollablePane( int type) {
	Type = type;

	switch( Type) {
		case SB_BOTH:
			Horz = new HScrollBar();
			Vert = new VScrollBar();
		break;
		case SB_VERT:
			Vert = new VScrollBar();
			break;
		case SB_HORZ:
			Horz = new HScrollBar();
			break;
	}

	setLayout( null);
	FBorder.setBorder( Border.BS_NONE);
}

public boolean addHotKey( int c, int mods, HotKeyProcessor p, Object sel){
	return BaseCv.addHotKey( c, mods, p, sel);
}

public boolean addHotKey ( int c, int mods, Object sel ) {
	return BaseCv.addHotKey( c, mods, null, sel);
}

public boolean canGetFocus(){
	if ( BaseCv != null)
		return BaseCv.canGetFocus();
	return false;
}

public int countLines(){
	if ( BaseCv.Lines == null)
		return 0;
	return BaseCv.Lines.size();
}

public void doLayout() {
	LayoutManager lm = getLayout();

	if ( lm != null)	lm.layoutContainer( this);
	else							updateBounds( 0, 0, 0, 0);
}

void drawFocusMark () {
	Graphics g = getGraphics();
	if ( g != null ) {
		drawFocusMark( g);
		g.dispose();
	}
}

protected void drawFocusMark ( Graphics g) {

	int b = Awt.ScrollBarWidth;
	int x = Width - b;
	int y = Height - b;

	// no place to paint focus mark ?
	if ( Type != SB_BOTH )
		return;
		
	if ( BaseCv.hasFocus() )
		GraphicsLib.drawFocusMark( g, x, y);
	else {
		Color c = getBackground();
		if ( c != null) {
			g.setColor( c);
			g.fillRect( x, y, b, b);
		}
	}

}

public void drawPane () {
	synchronized ( BaseCv ) {
		DrawGraphics = BaseCv.ResGraph;
		DrawRect = new Rectangle( BaseCv.XOffset, BaseCv.DrawRect.y,
		                          BaseCv.Width - BaseCv.XOffset - BaseCv.XBorder,
		                          BaseCv.DrawRect.height );
		OsDrawPane.notifyObservers();
		DrawRect = null;
		DrawGraphics = null;
	}
}

public ScrollableCanvas getCanvas(){
	return BaseCv;
}

public int getLineHeight () {
	return BaseCv.LineHeight;
}

public Menu getMenu(){
	return BaseCv.getMenu();
}

public Point getSBPositions () {
	int x = (Horz != null) ? Horz.getPosition() : 0;
	int y = (Vert != null) ? Vert.getPosition() : 0;
	return new Point( x, y); 
}

public int getTopIndex(){
	return (BaseCv != null) ? BaseCv.IdxTop : -1;
}

public int getType () {
	return Type;
}

public boolean hasFocus(){
	return BaseCv.hasFocus();
}

public boolean imageUpdate ( Image img, int infoflags, int x, int y,
		      int width, int height) {

	if ( infoflags == 3) //( width and height)
		Awt.DefToolkit.prepareImage( img, width, height, this);

	boolean ready = ( (infoflags & ImageObserver.ALLBITS) > 0);

	if (ready)
		BaseCv.redraw();

	return !ready;
}

public void registerObserver(){
	BaseCv.OsFocusLost.addObserver( this);
	BaseCv.OsFocusGot.addObserver( this);
}

public void setBorder ( int borderStyle ){
	// we don't have a border of our own, this is just forwarded to the canvas

	BaseCv.setBorder( borderStyle);
}

void setCanvas( ScrollableCanvas cv ) {
	BaseCv = cv;
	BaseCv.setFont( getFont() );
	FocusForward = cv;
	BaseCv.setBorder( DefaultBorderStyle);

	if ( InTitle != null) {
		Lab.setContents( InTitle.getText());
		Lab.setBorder( InTitle.getBorder());
		Lab.setFont( InTitle.getFont());
		Lab.setAlignment( InTitle.getHAlign(), InTitle.getVAlign());
		add( Lab);
	}
	add( BaseCv);
	if ( Horz != null)	add( Horz);
	if ( Vert != null)	add( Vert);

	OsCommand = BaseCv.OsCommand;
	registerObserver();
}

public static void setDefaultBorderStyle ( int style ) {
	DefaultBorderStyle = style;
}

public boolean setFocus(){
	if ( BaseCv != null)
		return BaseCv.setFocus();
	return false;
}

public synchronized void setFont ( Font f) {
	super.setFont( f);
	if ( BaseCv != null)
		BaseCv.setFont( f);
}

public synchronized void setForeground ( Color c) {
	if ( BaseCv != null)
		BaseCv.setForeground( c);
}

public void setLineHeight( int h){
	BaseCv.LineHeight = h;
}

public void setLineWidth( int w){
	BaseCv.LineWidth = w;
}

public void setMenu ( Menu m) {
	BaseCv.setMenu( m);
}

public void setSBPositions (Point p) {
	if ( Horz != null)
		Horz.setPosition( p.x);
	if ( Vert != null)
		Vert.setPosition( p.y);
}

public void setTitle ( Title t) {
	InTitle = t; //--- this will only be evaluated in setCanvas()
	if ( isShowing()) { //--- use text, font and alignment anyway
		Lab.setAlignment( t.getHAlign(), t.getVAlign());
		Lab.setFont( t.getFont());
		Lab.setContents( t.getText());
	}
}

public void unregisterObserver(){
	BaseCv.OsFocusLost.deleteObserver( this);
	BaseCv.OsFocusGot.deleteObserver( this);
}

public void update( Observable o, Object arg){
	if ( o == BaseCv.OsFocusLost )
		drawFocusMark();
	else if ( o == BaseCv.OsFocusGot )
		drawFocusMark();
}

protected void updateBounds( int x0, int y0, int x1, int y1) {

	// scrollbar bounds have to be updated before !!! canvas
	// because canvas posChanged() updates slider settings
	
	
	int yT = (InTitle != null) ? InTitle.getHeight() : 0;
	
	if ( InTitle != null) Lab.setBounds(x0, 0, Width, yT);

	int c1 = Width - BarWidth;
	int c2 = Height- BarWidth;

	switch( Type) {
		case SB_BOTH:
			Vert.setBounds( c1-x1, y0+yT, BarWidth, c2-y1-(y0+yT));
			Horz.setBounds( x0, c2-y1, c1-x1-x0, BarWidth);
			BaseCv.setBounds( x0, y0+yT, c1-BarDist-x1-x0, c2-BarDist-y1-(y0+yT));
			break;
		case SB_VERT:
			Vert.setBounds( c1-x1, y0+yT, BarWidth, Height-y1-(y0+yT));
			BaseCv.setBounds( x0, y0+yT, c1-BarDist-x1-x0, Height-y1-(y0+yT));
			break;
		case SB_HORZ:
			Horz.setBounds( x0, c2-y1, Width-x1-x0, BarWidth);
			BaseCv.setBounds( x0, y0+yT, Width-x1-x0, c2-BarDist-y1-(y0+yT));
			break;
		case SB_NONE:
			BaseCv.setBounds( x0, y0+yT, Width-x1-x0, Height-y1-(y0+yT));
			break;
	}
}
}
