package biss.awt;

import biss.Command;
import biss.CommandQueue;
import biss.ObserverSocket;
import java.awt.AWTEvent;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.EventQueue;
import java.awt.Font;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.peer.ActiveEvent;
import java.util.Observable;
import java.util.Observer;

/**
 * A SpinButton widget (TextField with two Buttons to circulate through
 * a list of predefined values)
 *
 * (C) 1996,97 BISS GmbH Germany, see file 'LICENSE.BISS-AWT' for details
 * @author J.H.Mehlitz
 */
public class SpinButton
  extends Compound
  implements FocusSetInterface, Observer, HotKeyProcessor, MouseListener
{
	int CurVal = 0;
	TextField Entry = new TextField();
	int Max = 100;
	int Min = 0;
	int ScrollOp = 0;
/**
 * triggered by spin-up transition (final spin position reached)
 * or the TextFields Enter key.
 * parameter: -
 */
	public ObserverSocket OsSpinChanged = new ObserverSocket( this);
/**
 * triggered by any contents change (spin operation or setSelection() ) 
 * parameter: -
 */
	public ObserverSocket OsValueChanged = new ObserverSocket( this);
	Object[] Values = null;
	int SpinStep = 1;
	SpinHandler Repeater = new SpinHandler();

class SpinHandler
  extends RepeatCommand
{
	int SpinCount;

public SpinHandler () {
	super( null, true);
}

public void actionNotify () {

	SpinCount++;

	if ( ScrollOp == 1){
		if ( CurVal < Max)
			spinUp( SpinStep);
		else stop();
	}
	else if ( ScrollOp == 2){
		if ( CurVal > Min)
			spinDown( SpinStep);
		else stop();
	}
	
	if ( Mouse.isAnyActive() )
		repeat();
	else
		SpinCount = 0;
}

public void stop () {
	if ( SpinCount == 0 )
		actionNotify();
	SpinCount = 0;
	super.stop();
}
}

public SpinButton(){
	FocusForward = Entry;
	Entry.setReadOnly( true);
	Entry.setRightAlign( true);

	Entry.setBorder( Border.BS_NONE);
	FBorder.setBorder( TextField.DefaultBorderStyle);

	setLayout( null);
	add( Entry);

	setCursor( Cursor.getPredefinedCursor( Awt.ActionCursor));
	registerObserver();
	registerHotKeys();
	addMouseListener( this);
}

public SpinButton ( Object[] values) {
	this();
	setContents( values);
}

public SpinButton ( int min, int max) {
	this();
	setRange( min, max);
}

public boolean atLimit() {
	if ( CurVal == Max) return true;
	if ( CurVal == Min) return true;
	return false;
}

public boolean canGetFocus(){
	return Entry.canGetFocus();
}

public void clear() {
	Entry.setContents( "");
	CurVal = 0;
}

public void doLayout(){
	int bw = FBorder.Ext;
	Entry.setBounds( bw, bw, Width-Awt.ScrollBarWidth-2-2*bw, Height-2*bw );
}

public int getMaximum() {
	return Max;
}

public int getMinimum() {
	return Min;
}

public Dimension getPreferredSize(){
	return new Dimension( 7*Awt.SysFontWidth, 3*Awt.SysFontHeight/2);
}

public int getSelection() {
	return CurVal;
}

public Object getSelectionValue() {
	if ( Values == null)
		return null;
	return Values[CurVal];
}

public int getSpinStep() {
	return SpinStep;
}

public int getValueIdx( Object obj) {
	if ( Values == null)
		return -1;
	for ( int i=0; i<Values.length; i++){
		if ( obj.equals( Values[i] ))
			return i;
	}
	return -1;
}

public Object[] getValues() {
	return Values;
}

public boolean hasFocus(){
	return Entry.hasFocus();
}

public void mouseClicked( MouseEvent e) {
	e.consume();
}

public void mouseEntered( MouseEvent e) {
	if ( Enabled)
		setCursor();
	e.consume();
}

public void mouseExited( MouseEvent e) {
	e.consume();
}

public void mousePressed( MouseEvent e) {
	e.consume();

	if ( ! Enabled)
		return;

	// for we have no GUIHandler, keep states consistent
	Mouse.down( e);
	GUIHandler.disposeSingleInstWnd( this);

	int x = e.getX();
	int y = e.getY();
	if ( x > Entry.Width + FBorder.Ext) {
		if ( y < FBorder.Ext+Entry.Height/2 ){
			if ( CurVal < Max)
				ScrollOp = 1;
		}
		else if ( CurVal > Min)
			ScrollOp = 2;
	}

	if ( ScrollOp > 0) {
		redrawSpins();
		Repeater.start();
	}
}

public void mouseReleased( MouseEvent e) {
	e.consume();
	
	if ( !Enabled)
		return;
		
	Mouse.reset();
	
	if ( ScrollOp > 0) {
		Repeater.stop();
		ScrollOp = 0;
		redrawSpins();
		OsSpinChanged.notifyObservers( this);
	}
}

public void processKey( Object sel) {
	if ( sel.equals( "spinUp"))
		spinUp( SpinStep);
	else if ( sel.equals( "spinDown"))
		spinDown( SpinStep);
}

public void redraw( Graphics g) {
	redrawSpins( g);
}

void redrawSpins () {
	Graphics g = getGraphics();
	if ( g != null ) {
		redrawSpins( g);
		g.dispose();
	}
}

void redrawSpins( Graphics g) {
	int x = Entry.Width + FBorder.Ext;
	int w = Awt.ScrollBarWidth + 2;

	GraphicsLib.paintSpinButton( g, x, FBorder.Ext, w, Entry.Height, ScrollOp);
}

void registerHotKeys() {
	Entry.addHotKey( Awt.SpinUpKeys[0], Awt.SpinUpKeys[1], this, "spinUp");
	Entry.addHotKey( Awt.SpinDownKeys[0], Awt.SpinDownKeys[1], this, "spinDown");
}

public void registerObserver() {
	Entry.OsFocusGot.addObserver( this);
	Entry.OsFocusLost.addObserver( this);
	Entry.OsEnter.addObserver( this);
}

public synchronized void setBackground( Color c){
	Entry.setBackground( c);
}

public void setContents( Object[] va){
	Values = va;
	Min    = 0;
	Max    = Values.length-1;
	CurVal = 0;

	setCurrentCont();
}

protected void setCurrentCont(){

	if ( Values != null)
		Entry.setContents( Values[CurVal].toString() );
	else
		Entry.setContents( Integer.toString( CurVal));
}

public void setEnabled ( boolean b, boolean hideIfDisabled) {
	if ( hideIfDisabled) {
		if   ( b ) setCurrentCont();
		else       Entry.setContents( " ");
	}
	setEnabled( b);
}

public boolean setFocus(){
	return Entry.setFocus();
}

public synchronized void setFont( Font f){
	Entry.setFont( f);
}

public synchronized void setForeground( Color c){
	Entry.setForeground( c);
}

public void setRange( int min, int max){
	Values = null;
	Min    = min;
	Max    = max;
	CurVal = min;
	
	setCurrentCont();
}

public void setRange( int min, int max, int cur){
	Values = null;
	Min    = min;
	Max    = max;
	CurVal = cur;

	Entry.setContents( Integer.toString( CurVal));  
}

public void setReadOnly( boolean state){
	Entry.setReadOnly( state);
}

public void setSelection( int val){

	if ( val > Max)       val = Max;
	else if ( val < Min)  val = Min;

	CurVal = val;

	setCurrentCont();

	OsValueChanged.notifyObservers();
}

public boolean setSelectionValue ( Object val) {
	if ( Values == null)
		return false;
	for ( int i=0; i<Values.length; i++) {
		if ( Values[i].equals( val)) {
			setSelection( i);
			return true;
		}
	}
	return false;
}

public void setSpinStep( int st){
	SpinStep = st;
}

public void setStepContents( int val){
	setSelection( val*SpinStep);
}

public void setStepRange( int min, int max){
	setRange( min*SpinStep, max*SpinStep);
}

public void spinDown( int steps) {
	setSelection( CurVal - steps);
}

public void spinUp( int steps) {
	setSelection( CurVal + steps);
}

public void update( Observable obs, Object arg) {
	if ( obs == Entry.OsFocusGot)
		OsFocusGot.notifyObservers();
	else if ( obs == Entry.OsFocusLost)
		OsFocusLost.notifyObservers();
	else if ( obs == Entry.OsEnter){
		try {			
			setSelection( Integer.parseInt( Entry.getContents() ) );
		}
		catch ( NumberFormatException nfe) {
			int vi = getValueIdx( Entry.getContents() );
			if ( vi > -1) CurVal = vi;
		}
		OsSpinChanged.notifyObservers();
	}
}

public void updateRange( int min, int max){
	Min    = min;
	Max    = max;

	if ( CurVal > max)	  CurVal = max;
	else if ( CurVal < min) CurVal = min;

	setCurrentCont();
}
}
