package java.awt;

import java.awt.event.FocusEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.PaintEvent;

/**
 * EventQueue is a platform-independent class that queues events, both
 * from the underlying peer classes and from trusted application classes.
 *
 * @version 1.16 05/02/97
 * @author Thomas Ball
 */
public class EventQueue
{
// From Thread.java



	private static int threadInitNumber;
	private EventQueueItem queue;
	private EventQueueItem paintQueue;

public EventQueue() {
	queue = null;
	paintQueue = null;
	String name = "AWT-EventQueue-" + nextThreadNum();
	new EventDispatchThread(name, this).start();
}

/* comment out until 1.2...
    /**
     * Dispatch an event to its source.
     *
     * @param theEvent an instance of java.awt.AWTEvent, or a
     * subclass of it.
     *
     *    protected void dispatchEvent(AWTEvent event) {
     *        Object src = event.getSource();
     *        if (src instanceof Component) {
     *            ((Component)src).dispatchEvent(event);
     *        } else if (src instanceof MenuComponent) {
     *            ((MenuComponent)src).dispatchEvent(event);
     *        }
     *    }
     */

    /*
     * Change the target of any pending KeyEvents because of a focus change.
     */
synchronized void changeKeyEventFocus(Object newSource) {
	EventQueueItem q = queue;
	for (; q != null; q = q.next) {
		if (q.event instanceof KeyEvent) {
			q.event.setSource(newSource);
		}
	}
}

/**
     * Remove an event from the queue and return it.  This method will
     * block until an event has been posted by another thread.
     * @return the next AWTEvent
     * @exception InterruptedException 
     *            if another thread has interrupted this thread.
     */
public synchronized AWTEvent getNextEvent() throws InterruptedException {

	EventQueueItem eqi;

	while (true) {
		if ( queue == null)
			wait( 30);
		if ( queue != null) {
			eqi = queue;
			queue = queue.next;
			return eqi.event;
		}
		else if ( paintQueue != null) {
			eqi = paintQueue;
			paintQueue = paintQueue.next;
			return eqi.event;
		}
//		else wait();  // might cause indefinite delays if no event traffic
	}

}

private static synchronized int nextThreadNum() {
	return threadInitNumber++;
}

/**
     * Return the first event without removing it.
     * @return the first AWTEvent
     */
public synchronized AWTEvent peekEvent() {
	if ( queue != null)
		return queue.event;
	if ( paintQueue != null)
		return paintQueue.event;
	return null;
}

/*
     * Return the first event of the specified type, if any.
     * @param id the id of the type of event desired.
     * @return the first AwtEvent of the requested type.
     */
public synchronized AWTEvent peekEvent(int id) {
	EventQueueItem q = queue;
	for (; q != null; q = q.next) {
		if (q.id == id) {
			return q.event;
		}
	}
	q = paintQueue;
	for (; q != null; q = q.next) {
		if (q.id == id) {
			return q.event;
		}
	}
	return null;
}

/**
     * Post a 1.1-style event to the EventQueue.
     *
     * @param theEvent an instance of java.awt.AWTEvent, or a
     * subclass of it.
     */
public synchronized void postEvent(AWTEvent theEvent) {
	EventQueueItem q;
	EventQueueItem eqi = new EventQueueItem(theEvent);

	if ( (eqi.id == PaintEvent.PAINT) || (eqi.id == PaintEvent.UPDATE) ){
		if ( paintQueue == null)
			paintQueue = eqi;
		else {
			for ( q = paintQueue; q != null; q = q.next){
				PaintEvent pe = (PaintEvent)q.event;
				if (pe.getSource() == theEvent.getSource()) {
					//System.out.println( "found similar");
					Rectangle rect = pe.getUpdateRect();
					Rectangle newRect = ((PaintEvent)theEvent).getUpdateRect();
					if ( (eqi.id == PaintEvent.UPDATE) && (q.id == PaintEvent.PAINT) ) {
						pe = (PaintEvent)theEvent;
						q.event = pe;
						q.id = PaintEvent.UPDATE;
						//System.out.println( "change paint to update");
					}
					if (!rect.equals(newRect))
						pe.setUpdateRect(rect.union(newRect));
					return;
				}
				if ( q.next == null)
					break;
			}
			//System.out.println( "adding: " + theEvent);
			q.next = eqi;
		}
	}
	else {
		if (queue == null) {
			queue = eqi;
			notifyAll();
		} else {
			for ( q = queue; q != null; q = q.next) {
				if (q.id == eqi.id){
					switch (q.id) {
					case Event.MOUSE_MOVE:
					case Event.MOUSE_DRAG:
						// New-style event id's never collide with 
						// old-style id's, so if the id's are equal, 
						// we can safely cast the queued event to be 
						// an old-style one.
						MouseEvent e = (MouseEvent)q.event;
						if (e.getSource() == ((MouseEvent)theEvent).getSource() &&
						    e.getModifiers() == ((MouseEvent)theEvent).getModifiers()) {
							q.event = eqi.event;// just replace old event
							return;
						}
						break;
					}
				}
				if (q.next == null)
					break;
			}
			q.next = eqi;
		}
	}
}

/*
     * Remove any pending events for the specified source object.
     * This method is normally called by the source's removeNotify method.
     */
synchronized void removeSourceEvents(Object source) {
	EventQueueItem entry = queue;
	EventQueueItem prev = null;
	while (entry != null) {
		if (entry.event.getSource().equals(source)) {
			if (prev == null) {
				queue = entry.next;
			} else {
				prev.next = entry.next;
			}
		}
		prev = entry;
		entry = entry.next;
	}

	entry = paintQueue;
	prev = null;
	while (entry != null) {
		if (entry.event.getSource().equals(source)) {
			if (prev == null) {
				paintQueue = entry.next;
			} else {
				prev.next = entry.next;
			}
		}
		prev = entry;
		entry = entry.next;
	}
}
}

class EventQueueItem
{
	AWTEvent event;
	int id;
	EventQueueItem next;

EventQueueItem(AWTEvent evt) {
	event = evt;
	id = evt.getID();
}
}
